from gunpowder.ext import malis

from .batch_filter import BatchFilter


class RenumberConnectedComponents(BatchFilter):
    """Find connected components of the same value, and replace each component
    with a new label.

    Args:

        labels (:class:`ArrayKey`):

            The label array to modify.
    """

    def __init__(self, labels):
        self.labels = labels

    def process(self, batch, request):
        components = batch.arrays[self.labels].data
        dtype = components.dtype
        simple_neighborhood = malis.mknhood3d()
        affinities_from_components = malis.seg_to_affgraph(
            components, simple_neighborhood
        )
        components, _ = malis.connected_components_affgraph(
            affinities_from_components, simple_neighborhood
        )
        batch.arrays[self.labels].data = components.astype(dtype)



class GenericJaxModel:
    """An interface for models to follow in order to train or predict. A model
    implementing this interface will need to contain not only the forward
    model but also loss and update fn. Some examples can be found in
    https://github.com/funkelab/funlib.learn.jax

    Args:

        is_training (``bool``):

            Indicating whether the model will be used for training
            or inferencing.
    """

    def __init__(self, is_training):
        pass

    def initialize(self, rng_key, inputs):
        """Initialize parameters for training.

        Args:

            rng_key (jax.random.PRNGKey):

                Seed for parameter initialization

            inputs (``dict``, ``string`` -> jnp.ndarray):

                Dictionary of inputs, provided to initialize parameters
                with the correct dimensions.

        Return:

            params (Any):

                Function should return an object encapsulating different
                parameters of the model.
        """
        raise RuntimeError("Unimplemented")

    def forward(self, params, inputs):
        """Run the forward model.

        Args:

            params (Any):

                Model parameters.

            inputs (``dict``, ``string`` -> jnp.ndarray):

                Dictionary of inputs.

        Return:

            outputs (``dict``, ``string`` -> jnp.ndarray):

                Dictionary of outputs.
        """
        raise RuntimeError("Unimplemented")

    def train_step(self, params, inputs, pmapped):
        """Run one iteration of training on the model.

        Args:

            params (Any):

                Model parameters.

            inputs (``dict``, ``string`` -> jnp.ndarray):

                Dictionary of inputs.

            pmapped (``bool``):

                Whether the function is run with `jax.pmap` or not.
                If pmapped across devices, the function should take care to
                synchronize gradients during the train step.
                The `axis_name` is set to the ``string`` "num_devices"

        Return:

            Tuple(new_params, outputs, loss)

                new_params (Any):

                    Updated model parameters.

                outputs (``dict``, ``string`` -> jnp.ndarray):

                    Dictionary of outputs.

                loss (Union[``float``, (``dict``, ``string`` -> ``float``)]):

                    Loss value of this iteration. Value can either be a single
                    ``float`` or a dictionary of multiple losses.
        """
        raise RuntimeError("Unimplemented")



import copy
import logging

import numpy as np
from scipy.spatial import KDTree

from gunpowder.array import Array
from gunpowder.array_spec import ArraySpec
from gunpowder.batch_request import BatchRequest
from gunpowder.coordinate import Coordinate
from gunpowder.graph_spec import GraphSpec
from gunpowder.morphology import enlarge_binary_map
from gunpowder.nodes.batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class AddVectorMap(BatchFilter):
    def __init__(
        self,
        src_and_trg_points,
        voxel_sizes,
        radius_phys,
        partner_criterion,
        stayinside_array_keys=None,
        pad_for_partners=(0, 0, 0),
    ):
        """Creates a vector map of shape [dim_vector, [shape_of_array]] (e.g. [3, 50,50,50] for an array of
            shape (50,50,50)) where every voxel which is close to a any source point location has a vector which points to
            one of the source point location's target location.
            Close to a point location in src_point includes all voxels which
                1) lie within distance radius_phys of the considered src point location
                2) (if stayinside_array_keys is not None), lie within the same segment as the src location in the
                    mask provided in stayinside_array_keys.
            The partner_criterion decides to which target location of the source point location that the vector of a
            voxel points (the different criterions are described below).

        Args:
            src_and_trg_points (dict):      Dictionary from :class:``ArrayKey`` of the vector map to be created
                                            to a tuple (:class:``GraphKey`` of the source points, :class:``GraphKey``
                                            of the target points) which define the source and target points.
            voxel_sizes (dict):             Dictionary from
                                            :class:``ArrayKey`` of the vector
                                            map to be created to a
                                            :class:`Coordinate` for the voxel
                                            size of the array.
            stayinside_array_keys (dict):  Dictionary from :class:``ArrayKey`` of the vector map to be created to
                                            :class:``ArrayKey`` of the stayinside_array.
                                            The stayinside_array is assumed to contain discrete objects labeled with
                                            different object ids. The object id at the specific source location is used
                                            to restrict the region where vectors are created around a source location.
                                            Voxels that are located outside of this object are set to zero.
                                            If stayinside_array_keys is None, all the voxels within distance
                                            radius_phys to the source location receive a vector.
            pad_for_partners (tuple):       n-dim tuple which defines padding of trg_points request in all dimensions
                                            (in world units).
                                            This might be used s.t. also partner locations which lie within the padded
                                            region, hence slightly outside of the vector map's roi, are considered.
            radius_phys (int):              Radius (in world units) to restrict region where vectors are created around
                                            a source location.
            partner_criterion(str):         'min_distance' or 'all'
                                            'min_distance': the vectors of all the voxels around a source location
                                            point to the same target location, namely the location which has the
                                            minimal distance to the considered source location.
                                            'all': all partner locations of a given source location are considered.
                                            The region around a source location is split up into (num_partners)-subregions
                                            where each voxel points to the target location for which this subregion
                                            is the closest.
        """

        self.array_to_src_trg_points = src_and_trg_points
        self.voxel_sizes = voxel_sizes
        self.array_keys_to_stayinside_array_keys = stayinside_array_keys
        self.pad_for_partners = pad_for_partners
        self.radius_phys = radius_phys
        self.partner_criterion = partner_criterion

    def setup(self):
        for (
            array_key,
            (src_points_key, trg_points_key),
        ) in self.array_to_src_trg_points.items():
            for points_key in [src_points_key, trg_points_key]:
                assert (
                    points_key in self.spec
                ), "Asked for {} in AddVectorMap from {}, where {} is not provided.".format(
                    array_key, points_key, points_key
                )
            neg_pad_for_partners = Coordinate(
                (self.pad_for_partners * np.asarray([-1])).tolist()
            )
            self.provides(
                array_key,
                ArraySpec(
                    roi=self.spec[src_points_key].roi.grow(
                        neg_pad_for_partners, neg_pad_for_partners
                    ),
                    voxel_size=self.voxel_sizes[array_key],
                    interpolatable=False,
                    dtype=np.float32,
                ),
            )

        self.enable_autoskip()

    def prepare(self, request):
        deps = BatchRequest()

        for (
            array_key,
            (src_points_key, trg_points_key),
        ) in self.array_to_src_trg_points.items():
            if array_key in request:
                # increase or set request for points to be array roi + padding for partners outside roi for target points
                deps[src_points_key] = GraphSpec(request[array_key].roi)
                padded_roi = request[array_key].roi.grow(
                    (self.pad_for_partners), (self.pad_for_partners)
                )
                deps[trg_points_key] = GraphSpec(padded_roi)

        for (
            array_key,
            stayinside_array_key,
        ) in self.array_keys_to_stayinside_array_keys.items():
            if array_key in request:
                deps[stayinside_array_key] = copy.deepcopy(request[array_key])

        return deps

    def process(self, batch, request):
        # create vector map and add it to batch
        for (
            array_key,
            (src_points_key, trg_points_key),
        ) in self.array_to_src_trg_points.items():
            if array_key in request:
                vector_map = self.__get_vector_map(
                    batch=batch, request=request, vector_map_array_key=array_key
                )
                spec = self.spec[array_key].copy()
                spec.roi = request[array_key].roi
                batch.arrays[array_key] = Array(data=vector_map, spec=spec)

    def __get_vector_map(self, batch, request, vector_map_array_key):
        src_points_key, trg_points_key = self.array_to_src_trg_points[
            vector_map_array_key
        ]
        dim_vectors = len(request[vector_map_array_key].roi.shape)
        voxel_size_vm = self.voxel_sizes[vector_map_array_key]
        offset_vector_map_phys = request[vector_map_array_key].roi.offset
        vector_map_total = np.zeros(
            (dim_vectors,) + (request[vector_map_array_key].roi.shape // voxel_size_vm),
            dtype=np.float32,
        )

        if batch.graphs[src_points_key].num_vertices() == 0:
            return vector_map_total

        for node in batch.graphs[src_points_key].nodes:
            if request[vector_map_array_key].roi.contains(Coordinate(node.location)):
                # get all partner locations which should be considered
                relevant_partner_loc = self.__get_relevant_partner_locations(
                    batch, node, trg_points_key
                )
                if len(relevant_partner_loc) > 0:
                    # get locations where to set vectors around source location
                    # (look only at region close to src location (to avoid np.nonzero() over entire array))
                    mask = self.__get_mask(
                        batch, request, vector_map_array_key, node.location
                    )
                    offset_vx_considered_mask = [
                        (
                            (
                                node.location[dim]
                                - self.radius_phys
                                - offset_vector_map_phys[dim]
                            )
                            // voxel_size_vm[dim]
                        )
                        for dim in range(dim_vectors)
                    ]
                    clipped_offset_vx_considered_mask = np.clip(
                        offset_vx_considered_mask, a_min=0, a_max=np.inf
                    )
                    slices = tuple(
                        slice(
                            int(np.max((0, offset_vx_considered_mask[dim]))),
                            int(
                                np.min(
                                    (
                                        offset_vx_considered_mask[dim]
                                        + (2 * self.radius_phys // voxel_size_vm[dim]),
                                        (mask.shape[dim]),
                                    )
                                )
                            ),
                        )
                        for dim in range(dim_vectors)
                    )
                    considered_region_mask = mask[slices]
                    locations_to_fill_vx = np.reshape(
                        np.nonzero(considered_region_mask), [dim_vectors, -1]
                    ).T
                    locations_to_fill_abs_phys = (
                        (
                            (locations_to_fill_vx + clipped_offset_vx_considered_mask)
                            * voxel_size_vm
                        )
                        + offset_vector_map_phys
                    ).tolist()

                    # check for target node with largest distance first and add vector pointing to it to vector_map_total
                    num_src_vectors_per_trg_loc = len(
                        locations_to_fill_abs_phys
                    ) // len(relevant_partner_loc)
                    if num_src_vectors_per_trg_loc > 0:
                        dist_to_locs = {}
                        for phys_loc in relevant_partner_loc:
                            dist_to_locs[np.linalg.norm(node.location - phys_loc)] = (
                                phys_loc
                            )
                        for nr, dist in enumerate(
                            reversed(np.sort(list(dist_to_locs.keys())))
                        ):
                            trg_loc_abs_phys = dist_to_locs[dist]
                            kdtree_locs_vector_map = KDTree(locations_to_fill_abs_phys)
                            if nr == len(relevant_partner_loc) - 1:
                                num_src_vectors_per_trg_loc = len(
                                    locations_to_fill_abs_phys
                                )
                            distances, ids = kdtree_locs_vector_map.query(
                                trg_loc_abs_phys, k=num_src_vectors_per_trg_loc
                            )

                            try:
                                len(ids)
                            except TypeError:
                                ids = [ids]

                            for src_voxel_id in ids:
                                # remove node from list which are taken as neighbors of THIS target location
                                neighbor_loc_abs_phys = kdtree_locs_vector_map.data[
                                    src_voxel_id
                                ]
                                locations_to_fill_abs_phys.remove(
                                    neighbor_loc_abs_phys.tolist()
                                )

                                # get vector for THIS neighbor to THIS target location, get its location and place it
                                vector = trg_loc_abs_phys - neighbor_loc_abs_phys
                                neighbor_loc_shifted_vx = (
                                    neighbor_loc_abs_phys - offset_vector_map_phys
                                ) // voxel_size_vm
                                for dim in range(dim_vectors):
                                    vector_map_total[dim][
                                        tuple([int(l)] for l in neighbor_loc_shifted_vx)
                                    ] = vector[dim]
        return vector_map_total

    def __get_relevant_partner_locations(self, batch, node, trg_points_key):
        # criterions: 'min_distance' or 'all'

        # get all partner locations
        all_partners_locations = []
        for partner_id in node.attrs["partner_ids"]:
            if batch.graphs[trg_points_key].contains(partner_id):
                all_partners_locations.append(
                    batch.graphs[trg_points_key].node(partner_id).location
                )

        # if only one partner location, return this one for any given criterion
        if len(all_partners_locations) <= 1:
            return all_partners_locations

        # return all partner locations
        elif self.partner_criterion == "all":
            return all_partners_locations

        # return partner with minimal euclidean distance to src_location
        elif self.partner_criterion == "min_distance":
            min_distance, stored_pos = np.inf, []
            for partner_loc in all_partners_locations:
                distance = np.linalg.norm(partner_loc - node.location)
                if distance < min_distance:
                    min_distance = distance.copy()
                    stored_pos = partner_loc.copy()
            return [stored_pos]

    def __get_mask(self, batch, request, vector_map_array_key, src_location):
        """create binary mask encoding where to place vectors for in region around considered src_location"""

        voxel_size = self.voxel_sizes[vector_map_array_key]

        offset_bm_phys = request[vector_map_array_key].roi.offset
        shape_bm_array_vx = request[vector_map_array_key].roi.shape // voxel_size
        binary_map = np.zeros(shape_bm_array_vx, dtype="uint8")

        if self.array_keys_to_stayinside_array_keys is None:
            mask = np.ones_like(binary_map)
        else:
            stayinside_array_key = self.array_keys_to_stayinside_array_keys[
                vector_map_array_key
            ]
            mask = batch.arrays[stayinside_array_key].data

        if mask.shape > binary_map.shape:
            # assumption: binary map is centered in the mask array
            padding = (np.asarray(mask.shape) - np.asarray(binary_map.shape)) / 2.0
            slices = [slice(np.floor(pad), -np.ceil(pad)) for pad in padding]
            mask = mask[slices]

        binary_map_total = np.zeros_like(binary_map)

        shifted_loc = src_location - np.asarray(offset_bm_phys)
        shifted_loc = shifted_loc.astype(np.int32) // voxel_size
        object_id = mask[tuple([loc] for loc in shifted_loc)][
            0
        ]  # 0 index, otherwise numpy array with single number

        binary_map[tuple([loc] for loc in shifted_loc)] = 1
        binary_map = enlarge_binary_map(
            binary_map, radius=self.radius_phys, voxel_size=voxel_size
        )
        binary_map[mask != object_id] = 0
        binary_map_total += binary_map
        binary_map.fill(0)
        binary_map_total[binary_map_total != 0] = 1

        return binary_map_total



import logging

import numpy as np
from scipy.ndimage import distance_transform_edt

from gunpowder.array import Array
from gunpowder.batch_request import BatchRequest
from gunpowder.nodes.batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class AddBoundaryDistanceGradients(BatchFilter):
    """Add an array with vectors pointing away from the closest boundary.

    The vectors are the spacial gradients of the distance transform, i.e., the
    distance to the boundary between labels or the background label (0).

    Args:

        label_array_key(:class:``ArrayKey``): The array to read the labels
            from.

        gradient_array_key(:class:``ArrayKey``): The array to generate
            containing the gradients.

        distance_array_key(:class:``ArrayKey``, optional): The array to
            generate containing the values of the distance transform.

        boundary_array_key(:class:``ArrayKey``, optional): The array to
            generate containing a boundary labeling. Note this array will be
            doubled as it encodes boundaries between voxels.

        normalize(string, optional): ``None``, ``'l1'``, or ``'l2'``. Specifies
            if and how to normalize the gradients.

        scale(string, optional): ``None`` or ``exp``. If ``exp``, distance
            gradients will be scaled by ``beta*e**(-d*alpha)``, where ``d`` is
            the distance to the boundary.

        scale_args(tuple, optional): For ``exp`` a tuple with the values of
            ``alpha`` and ``beta``.
    """

    def __init__(
        self,
        label_array_key,
        gradient_array_key,
        distance_array_key=None,
        boundary_array_key=None,
        normalize=None,
        scale=None,
        scale_args=None,
    ):
        self.label_array_key = label_array_key
        self.gradient_array_key = gradient_array_key
        self.distance_array_key = distance_array_key
        self.boundary_array_key = boundary_array_key
        self.normalize = normalize
        self.scale = scale
        self.scale_args = scale_args

    def setup(self):
        assert self.label_array_key in self.spec, (
            "Upstream does not provide %s needed by "
            "AddBoundaryDistanceGradients" % self.label_array_key
        )

        spec = self.spec[self.label_array_key].copy()
        spec.dtype = np.float32
        self.provides(self.gradient_array_key, spec)
        if self.distance_array_key is not None:
            self.provides(self.distance_array_key, spec)
        if self.boundary_array_key is not None:
            spec.voxel_size /= 2
            self.provides(self.boundary_array_key, spec)
        self.enable_autoskip()

    def prepare(self, request):
        deps = BatchRequest()
        deps[self.label_array_key] = request[self.gradient_array_key]

        return deps

    def process(self, batch, request):
        if self.gradient_array_key not in request:
            return

        labels = batch.arrays[self.label_array_key].data
        voxel_size = self.spec[self.label_array_key].voxel_size

        # get boundaries between label regions
        boundaries = self.__find_boundaries(labels)

        # mark boundaries with 0 (not 1)
        boundaries = 1.0 - boundaries

        if np.sum(boundaries == 0) == 0:
            # no boundary -- no distance to compute
            distances = np.zeros(labels.shape, dtype=np.float32)

        else:
            # get distances (voxel_size/2 because image is doubled)
            distances = distance_transform_edt(
                boundaries, sampling=tuple(float(v) / 2 for v in voxel_size)
            )
            distances = distances.astype(np.float32)

            # restore original shape
            downsample = (slice(None, None, 2),) * len(voxel_size)
            distances = distances[downsample]

            # set distances in background to 0
            distances[labels == 0] = 0

        gradients = np.asarray(np.gradient(distances, *voxel_size))

        # set gradients on background voxels to 0
        for d in range(len(voxel_size)):
            gradients[d, labels == 0] = 0

        if self.normalize is not None:
            self.__normalize(gradients, self.normalize)

        if self.scale is not None:
            self.__scale(gradients, distances, self.scale, self.scale_args)

        spec = self.spec[self.gradient_array_key].copy()
        spec.roi = request[self.gradient_array_key].roi
        batch.arrays[self.gradient_array_key] = Array(gradients, spec)

        if self.distance_array_key is not None and self.distance_array_key in request:
            batch.arrays[self.distance_array_key] = Array(distances, spec)

        if self.boundary_array_key is not None and self.boundary_array_key in request:
            # add one more face at each dimension, as boundary map has shape
            # 2*s - 1 of original shape s
            grown = np.ones(tuple(s + 1 for s in boundaries.shape))
            grown[tuple(slice(0, s) for s in boundaries.shape)] = boundaries
            spec.voxel_size = voxel_size / 2
            logger.debug("voxel size of boundary array: %s", spec.voxel_size)
            batch.arrays[self.boundary_array_key] = Array(grown, spec)

    def __find_boundaries(self, labels):
        # labels: 1 1 1 1 0 0 2 2 2 2 3 3       n
        # shift :   1 1 1 1 0 0 2 2 2 2 3       n - 1
        # diff  :   0 0 0 1 0 1 0 0 0 1 0       n - 1
        # bound.: 00000001000100000001000      2n - 1

        logger.debug("computing boundaries for %s", labels.shape)

        dims = len(labels.shape)
        in_shape = labels.shape
        out_shape = tuple(2 * s - 1 for s in in_shape)
        out_slices = tuple(slice(0, s) for s in out_shape)

        boundaries = np.zeros(out_shape, dtype=bool)

        logger.debug("boundaries shape is %s", boundaries.shape)

        for d in range(dims):
            logger.debug("processing dimension %d", d)

            shift_p = [slice(None)] * dims
            shift_p[d] = slice(1, in_shape[d])

            shift_n = [slice(None)] * dims
            shift_n[d] = slice(0, in_shape[d] - 1)

            diff = (labels[tuple(shift_p)] - labels[tuple(shift_n)]) != 0

            logger.debug("diff shape is %s", diff.shape)

            target = [slice(None, None, 2)] * dims
            target[d] = slice(1, out_shape[d], 2)

            logger.debug("target slices are %s", target)

            boundaries[tuple(target)] = diff

        return boundaries

    def __normalize(self, gradients, norm):
        dims = gradients.shape[0]

        if norm == "l1":
            factors = sum([np.abs(gradients[d]) for d in range(dims)])
        elif norm == "l2":
            factors = np.sqrt(sum([np.square(gradients[d]) for d in range(dims)]))
        else:
            raise RuntimeError("norm %s not supported" % norm)

        factors[factors < 1e-5] = 1
        gradients /= factors

    def __scale(self, gradients, distances, scale, scale_args):
        dims = gradients.shape[0]

        if scale == "exp":
            alpha, beta = self.scale_args
            factors = np.exp(-distances * alpha) * beta

        gradients *= factors



from .batch_filter import BatchFilter


class IntensityScaleShift(BatchFilter):
    """Scales the intensities of a batch by ``scale``, then adds ``shift``.

    Args:

        array (:class:`ArrayKey`):

            The key of the array to modify.

        scale (``float``):
        shift (``float``):

            The shift and scale to apply to ``array``.
    """

    def __init__(self, array, scale, shift):
        self.array = array
        self.scale = scale
        self.shift = shift

    def process(self, batch, request):
        if self.array not in batch.arrays:
            return

        raw = batch.arrays[self.array]
        raw.data = raw.data * self.scale + self.shift



import logging

from gunpowder.batch_request import BatchRequest
from gunpowder.profiling import Timing

from .batch_provider import BatchProvider

logger = logging.getLogger(__name__)


class BatchFilterError(Exception):
    def __init__(self, batch_filter, msg):
        self.batch_filter = batch_filter
        self.msg = msg

    def __str__(self):
        return f"Error in {self.batch_filter.name()}: {self.msg}"


class BatchFilter(BatchProvider):
    """Convenience wrapper for :class:`BatchProviders<BatchProvider>` with
    exactly one input provider.

    By default, a node of this class will expose the same :class:`ProviderSpec`
    as the upstream provider. You can modify the provider spec by calling
    :func:`provides` and :func:`updates` in :func:`setup`.

    Subclasses need to implement at least :func:`process` to modify a passed
    batch (downstream). Optionally, the following methods can be implemented:

        :func:`setup`

            Initialize this filter. Called after setup of the DAG. All upstream
            providers will be set up already.

        :func:`teardown`

            Destruct this filter, free resources, stop worker processes.

        :func:`prepare`

            Prepare for a batch request. Always called before each
            :func:`process`. Used to communicate dependencies.
    """

    @property
    def remove_placeholders(self):
        if not hasattr(self, "_remove_placeholders"):
            return False
        return self._remove_placeholders

    def get_upstream_provider(self):
        if len(self.get_upstream_providers()) != 1:
            raise BatchFilterError(
                self,
                "BatchFilters need to have exactly one upstream provider, "
                f"this one has {len(self.get_upstream_providers())}: "
                f"({[b.name() for b in self.get_upstream_providers()]}",
            )
        return self.get_upstream_providers()[0]

    def updates(self, key, spec):
        """Update an output provided by this :class:`BatchFilter`.

        Implementations should call this in their :func:`setup` method, which
        will be called when the pipeline is build.

        Args:

            key (:class:`ArrayKey` or :class:`GraphKey`):

                The array or point set key this filter updates.

            spec (:class:`ArraySpec` or :class:`GraphSpec`):

                The updated spec of the array or point set.
        """

        if key not in self.spec:
            raise BatchFilterError(
                self,
                f"BatchFilter {self} is trying to change the spec for {key}, "
                f"but {key} is not provided upstream. Upstream offers: "
                f"{self.get_upstream_provider().spec}",
            )
        self.spec[key] = spec.copy()
        self.updated_items.append(key)

        logger.debug("%s updates %s with %s" % (self.name(), key, spec))

    def enable_autoskip(self, skip=True):
        """Enable automatic skipping of this :class:`BatchFilter`, based on
        given :func:`updates` and :func:`provides` calls. Has to be called in
        :func:`setup`.

        By default, :class:`BatchFilters<BatchFilter>` are not skipped
        automatically, regardless of what they update or provide. If autskip is
        enabled, :class:`BatchFilters<BatchFilter>` will only be run if the
        request contains at least one key reported earlier with
        :func:`updates` or :func:`provides`.
        """

        self._autoskip_enabled = skip

    def _init_spec(self):
        # default for BatchFilters is to provide the same as upstream
        if not hasattr(self, "_spec") or self._spec is None:
            if len(self.get_upstream_providers()) != 0:
                self._spec = self.get_upstream_provider().spec.copy()
            else:
                self._spec = None

    def internal_teardown(self):
        logger.debug("Resetting spec of %s", self.name())
        self._spec = None
        self._updated_items = []

        self.teardown()

    @property
    def updated_items(self):
        """Get a list of the keys that are updated by this `BatchFilter`.

        This list is only available after the pipeline has been build. Before
        that, it is empty.
        """

        if not hasattr(self, "_updated_items"):
            self._updated_items = []

        return self._updated_items

    @property
    def autoskip_enabled(self):
        if not hasattr(self, "_autoskip_enabled"):
            self._autoskip_enabled = False

        return self._autoskip_enabled

    def provide(self, request):
        skip = self.__can_skip(request) or self.skip_node(request)

        timing_prepare = Timing(self, "prepare")
        timing_prepare.start()

        downstream_request = request.copy()

        if not skip:
            dependencies = self.prepare(request)
            if isinstance(dependencies, BatchRequest):
                upstream_request = request.update_with(dependencies)
            elif dependencies is None:
                upstream_request = request.copy()
            else:
                raise BatchFilterError(
                    self,
                    f"This BatchFilter returned a {type(dependencies)}! "
                    "Supported return types are: `BatchRequest` containing your exact "
                    "dependencies or `None`, indicating a dependency on the full request.",
                )
            self.remove_provided(upstream_request)
        else:
            upstream_request = request.copy()
        self.remove_provided(upstream_request)

        timing_prepare.stop()

        batch = self.get_upstream_provider().request_batch(upstream_request)

        timing_process = Timing(self, "process")
        timing_process.start()

        if not skip:
            if dependencies is not None:
                dependencies.remove_placeholders()
                node_batch = batch.crop(dependencies)
            else:
                node_batch = batch
            downstream_request.remove_placeholders()
            processed_batch = self.process(node_batch, downstream_request)
            if processed_batch is None:
                processed_batch = node_batch
            batch = batch.merge(processed_batch, merge_profiling_stats=False).crop(
                downstream_request
            )

        timing_process.stop()

        batch.profiling_stats.add(timing_prepare)
        batch.profiling_stats.add(timing_process)

        return batch

    def __can_skip(self, request):
        """Check if this filter needs to be run for the given request."""

        if not self.autoskip_enabled:
            return False

        for key, spec in request.items():
            if spec.placeholder:
                continue
            if key in self.provided_items:
                return False
            if key in self.updated_items:
                return False

        return True

    def skip_node(self, request):
        """To be implemented in subclasses.

        Skip a node if a condition is met. Can be useful if using a probability
        to determine whether to use an augmentation, for example.
        """
        pass

    def setup(self):
        """To be implemented in subclasses.

        Called during initialization of the DAG. Callees can assume that all
        upstream providers are set up already.

        In setup, call :func:`provides` or :func:`updates` to announce the
        arrays and points provided or changed by this node.
        """
        pass

    def prepare(self, request):
        """To be implemented in subclasses.

        Prepare for a batch request. Should return a :class:`BatchRequest` of
        needed dependencies. If None is returned, it will be assumed that all
        of request is needed.
        """
        return None

    def process(self, batch, request):
        """To be implemented in subclasses.

        Filter a batch, will be called after :func:`prepare`. Should return a
        :class:`Batch` containing modified Arrays and Graphs. Keys in the returned
        batch will replace the associated data in the original batch. If None is
        returned it is assumed that the batch has been modified in place. ``request``
        is the same as passed to :func:`prepare`, provided for convenience.

        Args:

            batch (:class:`Batch`):

                The batch received from upstream to be modified by this node.

            request (:class:`BatchRequest`):

                The request this node received. The updated batch should meet
                this request.
        """
        raise BatchFilterError(self, "does not implement 'process'")



import glob
import logging

import numpy as np

from gunpowder.array import Array
from gunpowder.array_spec import ArraySpec
from gunpowder.batch import Batch
from gunpowder.coordinate import Coordinate
from gunpowder.ext import pyklb
from gunpowder.profiling import Timing
from gunpowder.roi import Roi

from .batch_provider import BatchProvider

logger = logging.getLogger(__name__)


class KlbSource(BatchProvider):
    """A `KLB <https://bitbucket.org/fernandoamat/keller-lab-block-filetype>`_
    data source.

    Provides a single array from the given KLB dataset.

    Args:

        filename (``string``):

            The name of the KLB file. This string can be a glob expression
            (e.g., ``frame_*.klb``), in which case all files that match are
            sorted and stacked together to form an additional dimension (like
            time). The additional dimension will start at 0 and have a default
            voxel size of 1 (which can be overwritten using the ``array_spec``
            argument).

        array (:class:`ArrayKey`):

            ArrayKey that this source offers.

        array_spec (:class:`ArraySpec`, optional):

            An optional :class:`ArraySpec` to overwrite the array specs
            automatically determined from the KLB file. This is useful to set
            ``voxel_size``, for example. Only fields that are not ``None`` in
            the given :class:`ArraySpec` will be used.

        num_threads (``int``):

            An optional integer to pass to pyklb reader indicating the number
            of threads to use when reading klb files. Entering None causes
            uses the pyklb default, which now is based on the number of cores
            in the machine. This pyklb default is bad for jobs on the cluster that
            are limited to the number of cores requested, and 1 is recommended.

    """

    def __init__(self, filename, array, array_spec=None, num_threads=1):
        self.filename = filename
        self.array = array
        self.array_spec = array_spec
        self.num_threads = num_threads

        self.files = None
        self.ndims = None

    def setup(self):
        self.files = glob.glob(self.filename)
        self.files.sort()

        logger.info("Reading KLB headers of %d files...", len(self.files))
        headers = [pyklb.readheader(f) for f in self.files]
        spec = self.__read_spec(headers)

        self.provides(self.array, spec)

    def provide(self, request):
        timing = Timing(self)
        timing.start()

        batch = Batch()

        request_spec = request[self.array]

        logger.debug("Reading %s in %s...", self.array, request_spec.roi)

        voxel_size = self.spec[self.array].voxel_size

        # scale request roi to voxel units
        dataset_roi = request_spec.roi / voxel_size

        # shift request roi into dataset
        dataset_roi = dataset_roi - self.spec[self.array].roi.offset / voxel_size

        # create array spec
        array_spec = self.spec[self.array].copy()
        array_spec.roi = request_spec.roi

        # add array to batch
        batch.arrays[self.array] = Array(self.__read(dataset_roi), array_spec)

        logger.debug("done")

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch

    def __read_spec(self, headers):
        num_files = len(headers)
        assert num_files > 0
        common_header = headers[0]
        for header in headers:
            for attr in ["imagesize_tczyx", "pixelspacing_tczyx"]:
                assert (common_header[attr] == header[attr]).all(), (
                    "Headers of provided KLB files differ in attribute %s" % attr
                )
            assert (
                common_header["datatype"] == header["datatype"]
            ), "Headers of provided KLB files differ in attribute datatype"

        size = Coordinate(common_header["imagesize_tczyx"])
        voxel_size = Coordinate(common_header["pixelspacing_tczyx"])
        dtype = common_header["datatype"]

        # strip leading 1 dimensions
        while size[0] == 1 and len(size) > 1:
            size = size[1:]
            voxel_size = voxel_size[1:]

        # append num_files dimension
        if num_files > 1:
            size = (num_files,) + size
            voxel_size = (1,) + voxel_size

        dims = Coordinate(size)
        self.ndims = len(dims)

        if self.array_spec is not None:
            spec = self.array_spec
        else:
            spec = ArraySpec()

        if spec.voxel_size is None:
            spec.voxel_size = Coordinate(voxel_size)

        if spec.roi is None:
            offset = Coordinate((0,) * self.ndims)
            spec.roi = Roi(offset, dims * spec.voxel_size)

        if spec.dtype is not None:
            assert spec.dtype == dtype, (
                "dtype %s provided in array_specs for %s, but differs from "
                "dataset dtype %s"
                % (self.array_specs[self.array].dtype, self.array, dtype)
            )
        else:
            spec.dtype = dtype

        if spec.interpolatable is None:
            spec.interpolatable = np.issubdtype(spec.dtype, np.floating) or (
                spec.dtype == np.uint8
            )
            logger.warning(
                "WARNING: You didn't set 'interpolatable' for %s. "
                "Based on the dtype %s, it has been set to %s. "
                "This might not be what you want.",
                self.array,
                spec.dtype,
                spec.interpolatable,
            )

        return spec

    def __read(self, roi):
        if len(self.files) == 1:
            return self.__read_file(self.files[0], roi)

        else:
            file_indices = range(roi.begin[0], roi.end[0])

            file_roi = Roi(roi.begin[1:], roi.shape[1:])

            return np.array(
                [self.__read_file(self.files[i], file_roi) for i in file_indices]
            )

    def __read_file(self, filename, roi):
        # pyklb reads max-inclusive, gunpowder rois are max exclusive ->
        # subtract (1, 1, ...) from max coordinate
        if self.num_threads:
            return pyklb.readroi(
                filename,
                roi.begin,
                roi.end - (1,) * roi.dims,
                numthreads=self.num_threads,
            )
        else:
            return pyklb.readroi(filename, roi.begin, roi.end - (1,) * roi.dims)

    def __repr__(self):
        return self.filename



import logging

import numpy as np
from skimage.transform import rescale

from gunpowder.array import Array, ArrayKey
from gunpowder.batch import Batch
from gunpowder.batch_request import BatchRequest
from gunpowder.coordinate import Coordinate
from gunpowder.roi import Roi

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class Resample(BatchFilter):
    """Up- or downsample arrays in a batch to match a given voxel size. Note: Behavior is not a pixel-perfect copy of down/upsample nodes, because this node relies on skimage.transform.rescale to perform non-integer scaling factors.

    Args:

        source (:class:`ArrayKey`):

            The key of the array to resample.

        target_voxel_size (:class:`Coordinate`):

            The voxel size of the target.

        target (:class:`ArrayKey`):

            The key of the array to store the resampled ``source``.

        ndim (``int``, optional):

            Dimensionality of upsampling. This allows users to, for instance, specify against
            resampling in unused z-dimension when processing slices of anisotropic data.
            Default is to use the dimensionality of ``target_voxel_size``.

        interp_order (``int``, optional):

            The order of interpolation. The order has to be in the range 0-5:
                0: Nearest-neighbor
                1: Bi-linear (default)
                2: Bi-quadratic
                3: Bi-cubic
                4: Bi-quartic
                5: Bi-quintic

                Default is 0 if image.dtype is bool or interpolatable is False, and 1 otherwise.

    """

    def __init__(self, source, target_voxel_size, target, ndim=None, interp_order=None):
        assert isinstance(source, ArrayKey)
        assert isinstance(target, ArrayKey)

        self.source = source
        self.target_voxel_size = Coordinate(target_voxel_size)
        self.target = target
        if ndim is None:
            self.ndim = len(target_voxel_size)
        else:
            self.ndim = ndim
        self.interp_order = interp_order

    def setup(self):
        spec = self.spec[self.source].copy()
        source_voxel_size = self.spec[self.source].voxel_size
        spec.voxel_size = self.target_voxel_size
        self.pad = Coordinate(
            (0,) * (len(source_voxel_size) - self.ndim)
            + source_voxel_size[-self.ndim :]
        )

        spec.roi = spec.roi.snap_to_grid(
            np.lcm(source_voxel_size, self.target_voxel_size), mode="shrink"
        )

        self.provides(self.target, spec)
        self.enable_autoskip()

    def prepare(self, request):
        source_voxel_size = self.spec[self.source].voxel_size
        source_request = request[self.target].copy()
        source_request.voxel_size = source_voxel_size
        source_request.roi = source_request.roi.grow(
            self.pad, self.pad
        )  # Pad w/ 1 voxel per side for interpolation to avoid edge effects
        source_request.roi = source_request.roi.snap_to_grid(
            np.lcm(source_voxel_size, self.target_voxel_size), mode="grow"
        )
        source_request.roi = source_request.roi.intersect(
            self.spec[self.source].roi
        ).snap_to_grid(np.lcm(source_voxel_size, self.target_voxel_size), mode="shrink")

        deps = BatchRequest()
        deps[self.source] = source_request

        return deps

    def process(self, batch, request):
        source = batch.arrays[self.source]
        source_data = source.data
        source_voxel_size = self.spec[self.source].voxel_size

        scales = np.array(source_voxel_size) / np.array(self.target_voxel_size)
        scales = (1,) * (source_data.ndim - source_voxel_size.dims) + tuple(scales)

        if self.interp_order != 0 and (
            self.spec[self.source].interpolatable
            or self.spec[self.source].interpolatable is None
        ):
            resampled_data = rescale(
                source_data.astype(np.float32), scales, order=self.interp_order
            ).astype(source_data.dtype)
        else:  # Force nearest-neighbor interpolation for non-interpolatable arrays
            if self.interp_order is not None and self.interp_order != 0:
                logger.warning(
                    "Interpolation other than nearest-neighbor requested for non-interpolatable array. Using nearest-neighbor instead."
                )
            resampled_data = rescale(
                source_data.astype(np.float32), scales, order=0, anti_aliasing=False
            ).astype(source_data.dtype)

        target_spec = source.spec.copy()
        target_spec.roi = Roi(
            source.spec.roi.get_begin(),
            self.target_voxel_size
            * Coordinate(resampled_data.shape[-self.target_voxel_size.dims :]),
        )
        target_spec.voxel_size = self.target_voxel_size
        target_spec.dtype = resampled_data.dtype
        target_array = Array(resampled_data, target_spec)
        target_array.crop(request[self.target].roi)

        # create output array
        outputs = Batch()
        outputs.arrays[self.target] = target_array

        return outputs



import logging
import math
import random
import warnings

import numpy as np
from scipy import ndimage

from gunpowder.batch_request import BatchRequest
from gunpowder.coordinate import Coordinate
from gunpowder.ext import augment
from gunpowder.roi import Roi

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class ElasticAugment(BatchFilter):
    """(DEPRICATED) Elasticly deform a batch. Requests larger batches upstream to avoid data
    loss due to rotation and jitter.

    Args:

        control_point_spacing (``tuple`` of ``int``):

            Distance between control points for the elastic deformation, in
            voxels per dimension.

        jitter_sigma (``tuple`` of ``float``):

            Standard deviation of control point jitter distribution, in voxels
            per dimension.

        rotation_interval (``tuple`` of two ``floats``):

            Interval to randomly sample rotation angles from (0, 2PI).

        scale_interval (``tuple`` of two ``floats``):

            Interval to randomly sample scale factors from.

        prob_slip (``float``):

            Probability of a section to "slip", i.e., be independently moved in
            x-y.

        prob_shift (``float``):

            Probability of a section and all following sections to move in x-y.

        max_misalign (``int``):

            Maximal voxels to shift in x and y. Samples will be drawn
            uniformly. Used if ``prob_slip + prob_shift`` > 0.

        subsample (``int``):

            Instead of creating an elastic transformation on the full
            resolution, create one subsampled by the given factor, and linearly
            interpolate to obtain the full resolution transformation. This can
            significantly speed up this node, at the expense of having visible
            piecewise linear deformations for large factors. Usually, a factor
            of 4 can savely by used without noticable changes. However, the
            default is 1 (i.e., no subsampling).

        spatial_dims (``int``):

            The number of spatial dimensions in arrays. Spatial dimensions are
            assumed to be the last ones and cannot be more than 3 (default).
            Set this value here to avoid treating channels as spacial
            dimension. If, for example, your array is indexed as ``(c,y,x)``
            (2D plus channels), you would want to set ``spatial_dims=2`` to
            perform the elastic deformation only on x and y.

        use_fast_points_transform (``bool``):

            By solving for all of your points simultaneously with the following
            3 step proceedure:
            1) Rasterize nodes into numpy array
            2) Apply elastic transform to array
            3) Read out nodes via center of mass of transformed points
            You can gain substantial speed up as opposed to calculating the
            elastic transform for each point individually. However this may
            lead to nodes being lost during the transform.

        recompute_missing_points (``bool``):

            Whether or not to compute the elastic transform node wise for nodes
            that were lossed during the fast elastic transform process.

        p (``float``, optional):

            Probability applying the augmentation. Default is 1.0 (always
            apply). Should be a float value between 0 and 1. Lowering this value
            could be useful for computational efficiency and increasing
            augmentation space.
    """

    def __init__(
        self,
        control_point_spacing,
        jitter_sigma,
        rotation_interval,
        scale_interval=(1.0, 1.0),
        prob_slip=0,
        prob_shift=0,
        max_misalign=0,
        subsample=1,
        spatial_dims=3,
        use_fast_points_transform=False,
        recompute_missing_points=True,
        p=1.0,
    ):
        warnings.warn(
            "ElasticAugment is deprecated, please use the DeformAugment",
            DeprecationWarning,
        )

        self.control_point_spacing = control_point_spacing
        self.jitter_sigma = jitter_sigma
        self.rotation_start = rotation_interval[0]
        self.rotation_max_amount = rotation_interval[1] - rotation_interval[0]
        self.scale_min = scale_interval[0]
        self.scale_max = scale_interval[1]
        self.prob_slip = prob_slip
        self.prob_shift = prob_shift
        self.max_misalign = max_misalign
        self.subsample = subsample
        self.spatial_dims = spatial_dims
        self.use_fast_points_transform = use_fast_points_transform
        self.recompute_missing_points = recompute_missing_points
        self.p = p

    def skip_node(self, request):
        return random.random() > self.p

    def prepare(self, request):
        # get the voxel size
        self.voxel_size = self.__get_common_voxel_size(request)

        # get the total ROI of all requests
        total_roi = request.get_total_roi()
        logger.debug("total ROI is %s" % total_roi)

        # First, get the total ROI of the request in spatial dimensions only.
        # Channels and time don't matter. This is our master ROI.

        # get master ROI
        master_roi = Roi(
            total_roi.begin[-self.spatial_dims :],
            total_roi.shape[-self.spatial_dims :],
        )
        self.spatial_dims = master_roi.dims
        logger.debug("master ROI is %s" % master_roi)

        # make sure the master ROI aligns with the voxel size
        master_roi = master_roi.snap_to_grid(self.voxel_size, mode="grow")
        logger.debug("master ROI aligned with voxel size is %s" % master_roi)

        # get master roi in voxels
        master_roi_voxels = master_roi / self.voxel_size
        logger.debug("master ROI in voxels is %s" % master_roi_voxels)

        # Second, create a master transformation. This is a transformation that
        # covers all voxels of the all requested ROIs. The master transformation
        # is zero-based.

        # create a transformation with the size of the master ROI in voxels
        self.master_transformation = self.__create_transformation(
            master_roi_voxels.shape
        )

        # Third, crop out parts of the master transformation for each of the
        # smaller requested ROIs. Since these ROIs now have to align with the
        # voxel size (which for points does not have to be the case), we also
        # remember these smaller ROIs as target_rois in global world units.

        # crop the parts corresponding to the requested ROIs
        self.transformations = {}
        self.target_rois = {}
        deps = BatchRequest()
        for key, spec in request.items():
            spec = spec.copy()

            if spec.roi is None:
                continue

            target_roi = Roi(
                spec.roi.begin[-self.spatial_dims :],
                spec.roi.shape[-self.spatial_dims :],
            )
            logger.debug("downstream request spatial ROI for %s is %s", key, target_roi)

            # make sure the target ROI aligns with the voxel grid (which might
            # not be the case for points)
            target_roi = target_roi.snap_to_grid(self.voxel_size, mode="grow")
            logger.debug(
                "downstream request spatial ROI aligned with voxel grid for %s is %s",
                key,
                target_roi,
            )

            # remember target ROI (this is where the transformation will project
            # to)
            self.target_rois[key] = target_roi

            # get ROI in voxels
            target_roi_voxels = target_roi / self.voxel_size

            # get ROI relative to master ROI
            target_roi_in_master_roi_voxels = (
                target_roi_voxels - master_roi_voxels.begin
            )

            # crop out relevant part of transformation for this request
            transformation = np.copy(
                self.master_transformation[
                    (slice(None),) + target_roi_in_master_roi_voxels.get_bounding_box()
                ]
            )
            self.transformations[key] = transformation

            # get ROI of all voxels necessary to perfrom transformation
            #
            # for that we follow the same transformations to get from the
            # request ROI to the target ROI in master ROI in voxels, just in
            # reverse
            source_roi_in_master_roi_voxels = self.__get_source_roi(transformation)
            source_roi_voxels = (
                source_roi_in_master_roi_voxels + master_roi_voxels.begin
            )
            source_roi = source_roi_voxels * self.voxel_size

            # transformation is still defined on voxels relative to master ROI
            # in voxels (i.e., lowest source coordinate could be 5, but data
            # array we get later starts at 0).
            #
            # shift transformation to be indexed relative to beginning of
            # source_roi_voxels
            self.__shift_transformation(
                -source_roi_in_master_roi_voxels.begin, transformation
            )

            # update upstream request
            spec.roi = Roi(
                spec.roi.begin[: -self.spatial_dims]
                + source_roi.begin[-self.spatial_dims :],
                spec.roi.shape[: -self.spatial_dims]
                + source_roi.shape[-self.spatial_dims :],
            )

            deps[key] = spec

            logger.debug("upstream request roi for %s = %s" % (key, spec.roi))

        return deps

    def process(self, batch, request):
        for array_key, array in batch.arrays.items():
            if array_key not in self.target_rois:
                continue

            # for arrays, the target ROI and the requested ROI should be the
            # same in spatial coordinates
            assert (
                self.target_rois[array_key].begin
                == request[array_key].roi.begin[-self.spatial_dims :]
            ), "Target roi offset {} does not match request roi offset {}".format(
                self.target_rois[array_key].begin,
                request[array_key].roi.begin[-self.spatial_dims :],
            )

            assert (
                self.target_rois[array_key].shape
                == request[array_key].roi.shape[-self.spatial_dims :]
            ), "Target roi offset {} does not match request roi offset {}".format(
                self.target_rois[array_key].shape,
                request[array_key].roi.shape[-self.spatial_dims :],
            )

            # reshape array data into (channels,) + spatial dims
            shape = array.data.shape
            channel_shape = shape[: -self.spatial_dims]
            data = array.data.reshape((-1,) + shape[-self.spatial_dims :])

            # apply transformation on each channel
            data = np.array(
                [
                    augment.apply_transformation(
                        data[c],
                        self.transformations[array_key],
                        interpolate=self.spec[array_key].interpolatable,
                    )
                    for c in range(data.shape[0])
                ]
            )

            data_roi = request[array_key].roi / self.spec[array_key].voxel_size
            array.data = data.reshape(
                channel_shape + data_roi.shape[-self.spatial_dims :]
            )

            # restore original ROIs
            array.spec.roi = request[array_key].roi

        for graph_key, graph in batch.graphs.items():
            nodes = list(graph.nodes)

            if self.use_fast_points_transform:
                missed_nodes = self.__fast_point_projection(
                    self.transformations[graph_key],
                    nodes,
                    graph.spec.roi,
                    target_roi=self.target_rois[graph_key],
                )
                if not self.recompute_missing_points:
                    for node in set(missed_nodes):
                        graph.remove_node(node, retain_connectivity=True)
                    missed_nodes = []
            else:
                missed_nodes = nodes

            for node in missed_nodes:
                # logger.debug("projecting %s", node.location)

                # get location relative to beginning of upstream ROI
                location = node.location - graph.spec.roi.begin
                logger.debug("relative to upstream ROI: %s", location)

                # get spatial coordinates of node in voxels
                location_voxels = location[-self.spatial_dims :] / self.voxel_size

                # get projected location in transformation data space, this
                # yields voxel coordinates relative to target ROI
                projected_voxels = self.__project(
                    self.transformations[graph_key], location_voxels
                )

                logger.debug(
                    "projected in voxels, relative to target ROI: %s", projected_voxels
                )

                if projected_voxels is None:
                    logger.debug("node outside of target, skipping")
                    graph.remove_node(node, retain_connectivity=True)
                    continue

                # convert to world units (now in float again)
                projected = projected_voxels * np.array(self.voxel_size)

                logger.debug(
                    "projected in world units, relative to target ROI: %s", projected
                )

                # get global coordinates
                projected += np.array(self.target_rois[graph_key].begin)

                # update spatial coordinates of node location
                node.location[-self.spatial_dims :] = projected

                logger.debug("final location: %s", node.location)

                # finally, it can happen that a node no longer is contained in
                # the requested ROI (because larger ROIs than necessary have
                # been requested upstream)
                if not request[graph_key].roi.contains(node.location):
                    logger.debug("node outside of target, skipping")
                    graph.remove_node(node, retain_connectivity=True)
                    continue

            # restore original ROIs
            graph.spec.roi = request[graph_key].roi

    def __get_common_voxel_size(self, request):
        voxel_size = None
        prev = None
        for array_key in request.array_specs.keys():
            if voxel_size is None:
                voxel_size = self.spec[array_key].voxel_size[-self.spatial_dims :]
            elif self.spec[array_key].voxel_size is not None:
                assert (
                    voxel_size == self.spec[array_key].voxel_size[-self.spatial_dims :]
                ), (
                    "ElasticAugment can only be used with arrays of same voxel sizes, "
                    "but %s has %s, and %s has %s."
                    % (
                        array_key,
                        self.spec[array_key].voxel_size,
                        prev,
                        self.spec[prev].voxel_size,
                    )
                )
            prev = array_key

        if voxel_size is None:
            raise RuntimeError("voxel size must not be None")

        return Coordinate(voxel_size)

    def __create_transformation(self, target_shape):
        scale = self.scale_min + random.random() * (self.scale_max - self.scale_min)

        transformation = augment.create_identity_transformation(
            target_shape, subsample=self.subsample, scale=scale
        )
        if sum(self.jitter_sigma) > 0:
            transformation += augment.create_elastic_transformation(
                target_shape,
                self.control_point_spacing,
                self.jitter_sigma,
                subsample=self.subsample,
            )
        rotation = random.random() * self.rotation_max_amount + self.rotation_start
        if rotation != 0:
            transformation += augment.create_rotation_transformation(
                target_shape, rotation, subsample=self.subsample
            )

        if self.subsample > 1:
            transformation = augment.upscale_transformation(
                transformation, target_shape
            )

        if self.prob_slip + self.prob_shift > 0:
            self.__misalign(transformation)

        return transformation

    def __fast_point_projection(self, transformation, nodes, source_roi, target_roi):
        if len(nodes) < 1:
            return []
        # rasterize the points into an array
        ids, locs = zip(
            *[
                (
                    node.id,
                    (np.floor(node.location).astype(int) - source_roi.begin)
                    // self.voxel_size,
                )
                for node in nodes
                if source_roi.contains(node.location)
            ]
        )
        ids, locs = np.array(ids), tuple(zip(*locs))
        points_array = np.zeros(source_roi.shape / self.voxel_size, dtype=np.int64)
        points_array[locs] = ids

        # reshape array data into (channels,) + spatial dims
        shape = points_array.shape
        data = points_array.reshape((-1,) + shape[-self.spatial_dims :])

        # apply transformation on each channel
        data = np.array(
            [
                augment.apply_transformation(
                    data[c], transformation, interpolate="nearest"
                )
                for c in range(data.shape[0])
            ]
        )

        missing_points = []
        projected_locs = ndimage.center_of_mass(data > 0, data, ids)
        projected_locs = [
            np.array(loc[-self.spatial_dims :]) * self.voxel_size + target_roi.begin
            for loc in projected_locs
        ]
        node_dict = {node.id: node for node in nodes}
        for point_id, proj_loc in zip(ids, projected_locs):
            point = node_dict.pop(point_id)
            if not any([np.isnan(x) for x in proj_loc]):
                assert (
                    len(proj_loc) == self.spatial_dims
                ), "projected location has wrong number of dimensions: {}, expected: {}".format(
                    len(proj_loc), self.spatial_dims
                )
                point.location[-self.spatial_dims :] = proj_loc
            else:
                missing_points.append(point)
        for node in node_dict.values():
            missing_points.append(point)
        logger.debug(
            "{} of {} points lost in fast points projection".format(
                len(missing_points), len(ids)
            )
        )

        return missing_points

    def __project(self, transformation, location):
        """Find the projection of location given by transformation. Returns None
        if projection lies outside of transformation."""

        dims = len(location)

        # subtract location from transformation
        diff = transformation.copy()
        for d in range(dims):
            diff[d] -= location[d]

        # square
        diff2 = diff * diff

        # sum
        dist = diff2.sum(axis=0)

        # find grid point closes to location
        center_grid = Coordinate(np.unravel_index(dist.argmin(), dist.shape))
        center_source = self.__source_at(transformation, center_grid)

        logger.debug("projecting %s onto grid", location)
        logger.debug("grid shape: %s", transformation.shape[1:])
        logger.debug("grid projection: %s", center_grid)
        logger.debug("dist shape: %s", dist.shape)
        logger.debug("dist.argmin(): %s", dist.argmin())
        logger.debug("dist[argmin]: %s", dist[center_grid])
        logger.debug(
            "transform[argmin]: %s", transformation[(slice(None),) + center_grid]
        )
        logger.debug("min dist: %s", dist.min())
        logger.debug("center source: %s", center_source)

        # inspect grid edges incident to center_grid
        for d in range(dims):
            # nothing to do for dimensions without spatial extent
            if transformation.shape[1 + d] == 1:
                continue

            dim_vector = Coordinate(1 if dd == d else 0 for dd in range(dims))
            pos_grid = center_grid + dim_vector
            neg_grid = center_grid - dim_vector
            logger.debug("interpolating along %s", dim_vector)

            pos_u = -1
            neg_u = -1

            if pos_grid[d] < transformation.shape[1 + d]:
                pos_source = self.__source_at(transformation, pos_grid)
                logger.debug("pos source: %s", pos_source)
                pos_dist = pos_source[d] - center_source[d]
                loc_dist = location[d] - center_source[d]
                if pos_dist != 0:
                    pos_u = loc_dist / pos_dist
                else:
                    pos_u = 0

            if neg_grid[d] >= 0:
                neg_source = self.__source_at(transformation, neg_grid)
                logger.debug("neg source: %s", neg_source)
                neg_dist = neg_source[d] - center_source[d]
                loc_dist = location[d] - center_source[d]
                if neg_dist != 0:
                    neg_u = loc_dist / neg_dist
                else:
                    neg_u = 0

            logger.debug("pos u/neg u: %s/%s", pos_u, neg_u)

            # if a point only falls behind edges, it lies outside of the grid
            if pos_u < 0 and neg_u < 0:
                return None

        return np.array(center_grid, dtype=np.float32)

    def __source_at(self, transformation, index):
        """Read the source point of a transformation at index."""

        slices = (slice(None),) + tuple(slice(i, i + 1) for i in index)
        return transformation[slices].flatten()

    def __get_source_roi(self, transformation):
        dims = transformation.shape[0]

        # get bounding box of needed data for transformation
        bb_min = Coordinate(
            int(math.floor(transformation[d].min())) for d in range(dims)
        )
        bb_max = Coordinate(
            int(math.ceil(transformation[d].max())) + 1 for d in range(dims)
        )

        # create roi sufficiently large to feed transformation
        source_roi = Roi(bb_min, bb_max - bb_min)

        return source_roi

    def __shift_transformation(self, shift, transformation):
        for d in range(transformation.shape[0]):
            transformation[d] += shift[d]

    def __misalign(self, transformation):
        assert (
            transformation.shape[0] == 3
        ), "misalign can only be applied to 3D volumes"

        num_sections = transformation[0].shape[0]

        shifts = [Coordinate((0, 0, 0))] * num_sections
        for z in range(num_sections):
            r = random.random()

            if r <= self.prob_slip:
                shifts[z] = self.__random_offset()

            elif r <= self.prob_slip + self.prob_shift:
                offset = self.__random_offset()
                for zp in range(z, num_sections):
                    shifts[zp] += offset

        logger.debug("misaligning sections with " + str(shifts))

        dims = 3
        bb_min = tuple(int(math.floor(transformation[d].min())) for d in range(dims))
        bb_max = tuple(int(math.ceil(transformation[d].max())) + 1 for d in range(dims))
        logger.debug("min/max of transformation: " + str(bb_min) + "/" + str(bb_max))

        for z in range(num_sections):
            transformation[1][z, :, :] += shifts[z][1]
            transformation[2][z, :, :] += shifts[z][2]

        bb_min = tuple(int(math.floor(transformation[d].min())) for d in range(dims))
        bb_max = tuple(int(math.ceil(transformation[d].max())) + 1 for d in range(dims))
        logger.debug(
            "min/max of transformation after misalignment: "
            + str(bb_min)
            + "/"
            + str(bb_max)
        )

    def __random_offset(self):
        return Coordinate(
            (0,)
            + tuple(
                self.max_misalign - random.randint(0, 2 * int(self.max_misalign))
                for d in range(2)
            )
        )



import copy
import logging
from copy import deepcopy

import numpy as np

from .freezable import Freezable

logger = logging.getLogger(__name__)


class Array(Freezable):
    """A numpy array with a specification describing the data.

    Args:

        data (array-like):

            The data to be stored in the array. Will be converted to a numpy
            array, if necessary.

        spec (:class:`ArraySpec`, optional):

            A spec describing the data.

        attrs (``dict``, optional):

            Optional attributes to describe this array.
    """

    def __init__(self, data, spec=None, attrs=None):
        self.spec = spec.copy()
        self.data = np.asarray(data)
        self.attrs = attrs

        if attrs is None:
            self.attrs = {}

        if spec is not None and spec.roi is not None and spec.voxel_size is not None:
            for d in range(len(spec.voxel_size)):
                assert (
                    spec.voxel_size[d] * data.shape[-spec.roi.dims + d]
                    == spec.roi.shape[d]
                ), "ROI %s does not align with voxel size %s * data shape %s" % (
                    spec.roi,
                    spec.voxel_size,
                    data.shape,
                )
                if spec.roi.offset[d] is not None:
                    assert (
                        spec.roi.offset[d] % spec.voxel_size[d] == 0
                    ), "ROI offset %s must be a multiple of voxel size %s" % (
                        spec.roi.offset,
                        spec.voxel_size,
                    )

        if spec.dtype is not None:
            assert (
                data.dtype == spec.dtype
            ), "data dtype %s does not match spec dtype %s" % (data.dtype, spec.dtype)

        self.freeze()

    def crop(self, roi, copy=True):
        """Create a cropped copy of this Array.

        Args:

            roi (:class:`Roi`):

                ROI in world units to crop to.

            copy (``bool``):

                Make a copy of the data.
        """

        assert self.spec.roi.contains(
            roi
        ), "Requested crop ROI (%s) doesn't fit in array (%s)" % (roi, self.spec.roi)

        if self.spec.roi == roi and not copy:
            return self

        voxel_size = self.spec.voxel_size
        data_roi = (roi - self.spec.roi.offset) / voxel_size
        slices = data_roi.get_bounding_box()

        while len(slices) < len(self.data.shape):
            slices = (slice(None),) + slices

        data = self.data[slices]
        if copy:
            data = np.array(data)

        spec = self.spec.copy()
        attrs = deepcopy(self.attrs)
        spec.roi = roi.copy()
        return Array(data, spec, attrs)

    def merge(self, array, copy_from_self=False, copy=False):
        """Merge this array with another one. The resulting array will have the
        size of the larger one, with values replaced from ``array``.

        This only works if one of the two arrays is contained in the other. In
        this case, ``array`` will overwrite values in ``self`` (unless
        ``copy_from_self`` is set to ``True``).

        A copy will only be made if necessary or ``copy`` is set to ``True``.
        """
        # It is unclear how to merge arrays in all cases. Consider a 10x10 array,
        # you crop out a 5x5 area, do a shift augment, and attempt to merge.
        # What does that mean? specs have changed. It should be a new key.
        raise NotImplementedError("Merge function should not be used!")

        self_roi = self.spec.roi
        array_roi = array.spec.roi

        assert self_roi.contains(array_roi) or array_roi.contains(
            self_roi
        ), "Can not merge arrays that are not contained in each other."

        assert (
            self.spec.voxel_size == array.spec.voxel_size
        ), "Can not merge arrays with different voxel sizes."

        # make sure self contains array
        if not self_roi.contains(array_roi):
            return array.merge(self, not copy_from_self, copy)

        # -> here we know that self contains array

        # simple case, self overwrites all of array
        if copy_from_self:
            return self if not copy else self.copy()

        # -> here we know that copy_from_self == False

        # simple case, ROIs are the same
        if self_roi == array_roi:
            return array if not copy else array.copy()

        # part of self have to be replaced, a copy is needed
        merged = self.copy()

        voxel_size = self.spec.voxel_size
        data_roi = (array_roi - self_roi.offset) / voxel_size
        slices = data_roi.get_bounding_box()

        while len(slices) < len(self.data.shape):
            slices = (slice(None),) + slices

        merged.data[slices] = array.data

        return merged

    def __repr__(self):
        return str(self.spec)

    def copy(self):
        """Create a copy of this array."""
        return copy.deepcopy(self)


class ArrayKey(Freezable):
    """A key to identify arrays in requests, batches, and across nodes.

    Used as key in :class:`BatchRequest` and :class:`Batch` to retrieve array
    specs or arrays.

    Args:

        identifier (``string``):

            A unique, human readable identifier for this array key. Will be
            used in log messages and to look up arrays in requests and batches.
            Should be upper case (like ``RAW``, ``GT_LABELS``). The identifier
            is unique: Two array keys with the same identifier will refer to
            the same array.
    """

    def __init__(self, identifier):
        self.identifier = identifier
        self.hash = hash(identifier)
        self.freeze()
        logger.debug("Registering array key %s", self)
        setattr(ArrayKeys, self.identifier, self)

    def __eq__(self, other):
        return hasattr(other, "identifier") and self.identifier == other.identifier

    def __hash__(self):
        return self.hash

    def __repr__(self):
        return self.identifier


class ArrayKeys:
    """Convenience access to all created :class:``ArrayKey``s. A key generated
    with::

        raw = ArrayKey('RAW')

    can be retrieved as::

        ArrayKeys.RAW
    """

    pass



import copy

from .coordinate import Coordinate
from .freezable import Freezable


class ArraySpec(Freezable):
    """Contains meta-information about an array. This is used by
    :class:`BatchProviders<BatchProvider>` to communicate the arrays they
    offer, as well as by :class:`Arrays<Array>` to describe the data they
    contain.

    Attributes:

        roi (:class:`Roi`):

            The region of interested represented by this array spec. Can be
            ``None`` for nonspatial arrays or to indicate the true value is unknown.

        voxel_size (:class:`Coordinate`):

            The size of the spatial axises in world units. Can be ``None`` for
            nonspatial arrays or to indicate the true value is unknown.

        interpolatable (``bool``):

            Whether the values of this array can be interpolated.

        nonspatial (``bool``, optional):

            If set, this array does not represent spatial data (e.g., a list of
            labels for samples in a batch). ``roi`` and ``voxel_size`` have to
            be ``None``. No consistency checks will be performed.

        dtype (``np.dtype``):

            The data type of the array.
    """

    def __init__(
        self,
        roi=None,
        voxel_size=None,
        interpolatable=None,
        nonspatial=False,
        dtype=None,
        placeholder=False,
    ):
        self.roi = roi
        self.voxel_size = None if voxel_size is None else Coordinate(voxel_size)
        self.interpolatable = interpolatable
        self.nonspatial = nonspatial
        self.dtype = dtype
        self.placeholder = placeholder

        if nonspatial:
            assert roi is None, "Non-spatial arrays can not have a ROI"
            assert voxel_size is None, "Non-spatial arrays can not have a voxel size"

        self.freeze()

    def update_with(self, spec):
        if self.roi is not None and spec.roi is not None:
            self.roi = self.roi.union(spec.roi)
        elif spec.roi is not None:
            self.roi = spec.roi

        if spec.voxel_size is not None:
            self.voxel_size = spec.voxel_size

        if spec.interpolatable is not None:
            self.interpolatable = spec.interpolatable

        if spec.nonspatial is not None:
            self.nonspatial = spec.nonspatial

        if spec.dtype is not None:
            self.dtype = spec.dtype

        if spec.placeholder is not None:
            self.placeholder = spec.placeholder

    def copy(self):
        """Create a copy of this spec."""
        return copy.deepcopy(self)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __repr__(self):
        r = ""
        r += "ROI: " + str(self.roi) + ", "
        r += "voxel size: " + str(self.voxel_size) + ", "
        r += "interpolatable: " + str(self.interpolatable) + ", "
        r += "non-spatial: " + str(self.nonspatial) + ", "
        r += "dtype: " + str(self.dtype) + ", "
        r += "placeholder: " + str(self.placeholder)
        return r



import ctypes
import logging
import multiprocessing as mp
from functools import reduce
from operator import mul

import numpy as np

from gunpowder.array import Array, ArrayKey
from gunpowder.ext import tensorflow as tf
from gunpowder.nodes.generic_predict import GenericPredict
from gunpowder.tensorflow.local_server import LocalServer

logger = logging.getLogger(__name__)


class Predict(GenericPredict):
    """Tensorflow implementation of :class:`gunpowder.nodes.Predict`.

    Args:

        checkpoint (``string``):

            Basename of a tensorflow checkpoint storing the tensorflow graph
            and associated tensor values and metadata, as created by
            :class:`gunpowder.nodes.Train`, for example.

        inputs (``dict``, ``string`` -> :class:`ArrayKey`):

            Dictionary from the names of input tensors in the network to
            array keys.

        outputs (``dict``, ``string`` -> :class:`ArrayKey`):

            Dictionary from the names of output tensors in the network to array
            keys. New arrays will be generated by this node for each entry (if
            requested downstream).

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            Used to set the specs of generated arrays (``outputs``). This is
            useful to set the ``voxel_size``, for example, if they differ from
            the voxel size of the input arrays. Only fields that are not
            ``None`` in the given :class:`ArraySpec` will be used.

        graph: (``string``, optional):

            An optional path to a tensorflow computation graph that should be
            used for prediction. The checkpoint is used to restore the values
            of matching variable names in the graph. Note that the graph
            specified here can differ from the one associated to the
            checkpoint.

        skip_empty (``bool``, optional):

            Skip prediction, if all inputs are empty (contain only 0). In this
            case, outputs are simply set to 0.

        max_shared_memory (``int``, optional):

            The maximal amount of shared memory in bytes to allocate to send
            batches to the GPU processes. Defaults to 1GB.
    """

    def __init__(
        self,
        checkpoint,
        inputs,
        outputs,
        array_specs=None,
        graph=None,
        skip_empty=False,
        max_shared_memory=1024 * 1024 * 1024,
    ):
        super(Predict, self).__init__(inputs, outputs, array_specs)

        self.checkpoint = checkpoint
        self.meta_graph = graph
        self.session = None
        self.graph = None
        self.skip_empty = skip_empty

        self.manager = mp.Manager()
        self.max_shared_memory = max_shared_memory
        self.shared_input_array_config = self.manager.dict()
        self.shared_output_array_config = self.manager.dict()
        self.shared_input_arrays = {}
        self.shared_output_arrays = {}
        self.shared_input_memory = mp.RawArray(ctypes.c_float, self.max_shared_memory)
        self.shared_output_memory = mp.RawArray(ctypes.c_float, self.max_shared_memory)

        self.send_lock = mp.Lock()
        self.receive_lock = mp.Lock()
        self.predict_process_initialized = mp.Event()
        self.worker_sent_inputs = mp.Event()
        self.predict_received_inputs = mp.Event()
        self.predict_sent_outputs = mp.Event()

        self.predict_process = mp.Process(target=self.__predict)
        self.predict_process_crashed = mp.Value("i", False)
        self.predict_process.start()
        self.predict_process_initialized.wait()

    def predict(self, batch, request):
        if not self.shared_output_arrays:
            self.__init_shared_output_arrays()

        if self.skip_empty:
            can_skip = True
            for array_key in self.inputs.values():
                if batch[array_key].data.any():
                    can_skip = False
                    break

            if can_skip:
                logger.info(f"Skipping batch for request: {request} (all inputs are 0)")

                for name, array_key in self.outputs.items():
                    shape = self.shared_output_arrays[name].shape
                    dtype = self.shared_output_arrays[name].dtype

                    spec = self.spec[array_key].copy()
                    spec.roi = request[array_key].roi.copy()
                    batch.arrays[array_key] = Array(np.zeros(shape, dtype=dtype), spec)

                return

        logger.debug(f"predicting for request: {request}")

        output_tensors = self.__collect_outputs(request)
        input_data = self.__collect_provided_inputs(batch)

        self.send_lock.acquire()

        if not self.shared_input_arrays:
            if not self.shared_input_array_config:
                self.__create_shared_input_array_config(batch, request)
            self.__init_shared_input_arrays()

        self.__write_inputs_to_shared(input_data)
        self.worker_sent_inputs.set()

        self.receive_lock.acquire()

        self.predict_received_inputs.wait()
        self.__check_background_process([self.receive_lock, self.send_lock])

        self.predict_received_inputs.clear()
        self.send_lock.release()

        self.predict_sent_outputs.wait()

        self.predict_sent_outputs.clear()

        output_data = self.__read_outputs_from_shared(output_tensors)

        self.receive_lock.release()

        for array_key in output_tensors:
            spec = self.spec[array_key].copy()
            spec.roi = request[array_key].roi
            batch.arrays[array_key] = Array(output_data[array_key], spec)

        logger.debug("predicted")

    def __predict(self):
        """The background predict process."""

        try:
            # TODO: is the server still needed?
            target = LocalServer.get_target()
            logger.info("Initializing tf session, connecting to %s...", target)

            self.graph = tf.Graph()
            self.session = tf.Session(target=target, graph=self.graph)

            with self.graph.as_default():
                self.__read_checkpoint()

            if not self.shared_output_arrays:
                if not self.shared_output_array_config:
                    self.__create_shared_output_array_config()
                self.__init_shared_output_arrays()

            # from now on it is save to access the shared array configuration
            self.predict_process_initialized.set()

            # loop predict
            while True:
                # wait for inputs
                self.worker_sent_inputs.wait()
                self.worker_sent_inputs.clear()

                if not self.shared_input_arrays:
                    self.__init_shared_input_arrays()

                # read inputs
                input_data = self.__read_inputs_from_shared()
                self.predict_received_inputs.set()

                # compute outputs
                output_data = self.session.run(
                    {t: t for t in self.outputs.keys()}, feed_dict=input_data
                )

                # write outputs
                self.__write_outputs_to_shared(output_data)
                self.predict_sent_outputs.set()

        except Exception as e:
            self.predict_process_crashed.value = True

            # release locks and events
            self.predict_process_initialized.set()
            self.worker_sent_inputs.clear()
            self.predict_received_inputs.set()
            self.predict_sent_outputs.set()
            raise e

    def teardown(self):
        self.predict_process.terminate()
        self.predict_process.join()

    def __check_background_process(self, locks=[]):
        if self.predict_process_crashed.value:
            # release all locks before raising exception
            for l in locks:
                l.release()
            raise RuntimeError("Background process died.")

    def __read_checkpoint(self):
        # read the graph associated to the checkpoint
        if self.meta_graph is None:
            meta_graph_file = self.checkpoint + ".meta"
        # read alternative, custom graph
        else:
            meta_graph_file = self.meta_graph

        logger.info(
            "Reading graph from %s and weights from %s...",
            meta_graph_file,
            self.checkpoint,
        )

        saver = tf.train.import_meta_graph(meta_graph_file, clear_devices=True)

        # restore variables from checkpoint
        saver.restore(self.session, self.checkpoint)

    def __collect_outputs(self, request=None):
        """Get a dict:

            array key: tensor name

        If request is not None, return only outputs that are in request.
        """

        array_outputs = {}

        for tensor_name, array_key in self.outputs.items():
            if request is None or array_key in request:
                array_outputs[array_key] = tensor_name

        return array_outputs

    def __collect_provided_inputs(self, batch):
        """Get a dict:

        tensor name: ndarray
        """

        inputs = {}

        for input_name, input_key in self.inputs.items():
            if isinstance(input_key, ArrayKey):
                if input_key in batch.arrays:
                    inputs[input_name] = batch.arrays[input_key].data
                else:
                    logger.warn(
                        "batch does not contain %s, input %s will not be set",
                        input_key,
                        input_name,
                    )
            elif isinstance(input_key, np.ndarray):
                inputs[input_name] = input_key
            elif isinstance(input_key, str):
                inputs[input_name] = getattr(batch, input_key)
            else:
                raise Exception(
                    "Unknown network input key {}, can't be given to network".format(
                        input_key
                    )
                )

        return inputs

    def __create_shared_input_array_config(self, batch, request):
        """Store the shared array config in a shared dictionary. Should be run
        once by the first worker to submit a batch."""

        begin = 0
        for name, array_key in self.inputs.items():
            shape = batch[array_key].data.shape
            size = reduce(mul, shape, 1)
            dtype = batch[array_key].data.dtype

            self.shared_input_array_config[name] = (begin, size, shape, dtype)

            begin += size * np.dtype(dtype).itemsize
            assert (
                begin <= self.max_shared_memory
            ), "The input arrays exceed the max_shared_memory"

    def __create_shared_output_array_config(self):
        """To be called by predict process."""

        begin = 0
        for name, array_key in self.outputs.items():
            tensor = self.graph.get_tensor_by_name(name)
            shape = tensor.get_shape().as_list()
            size = reduce(mul, shape, 1)
            dtype = tensor.dtype.as_numpy_dtype

            self.shared_output_array_config[name] = (begin, size, tuple(shape), dtype)

            begin += size * np.dtype(dtype).itemsize
            assert (
                begin <= self.max_shared_memory
            ), "The output arrays exceed the max_shared_memory"

    def __init_shared_input_arrays(self):
        """Assign the shared memory to numpy arrays."""

        for name, (begin, size, shape, dtype) in self.shared_input_array_config.items():
            self.shared_input_arrays[name] = np.frombuffer(
                self.shared_input_memory, dtype=dtype, offset=begin, count=size
            ).reshape(shape)

    def __init_shared_output_arrays(self):
        """Assign the shared memory to numpy arrays."""

        for name, (
            begin,
            size,
            shape,
            dtype,
        ) in self.shared_output_array_config.items():
            self.shared_output_arrays[name] = np.frombuffer(
                self.shared_output_memory, dtype=dtype, offset=begin, count=size
            ).reshape(shape)

    def __write_inputs_to_shared(self, input_data):
        for tensor_name, data in input_data.items():
            self.shared_input_arrays[tensor_name][:] = data

    def __read_inputs_from_shared(self):
        return {
            tensor_name: self.shared_input_arrays[tensor_name].copy()
            for tensor_name in self.inputs.keys()
        }

    def __write_outputs_to_shared(self, output_data):
        for tensor_name, data in output_data.items():
            self.shared_output_arrays[tensor_name][:] = data

    def __read_outputs_from_shared(self, output_tensors):
        return {
            array_key: self.shared_output_arrays[tensor_name].copy()
            for array_key, tensor_name in output_tensors.items()
        }



import logging
from typing import Any, Dict, Optional, Union

import numpy as np

from gunpowder.array import Array, ArrayKey
from gunpowder.array_spec import ArraySpec
from gunpowder.ext import NoSuchModule, tensorboardX, torch
from gunpowder.nodes.generic_train import GenericTrain

logger = logging.getLogger(__name__)


class Train(GenericTrain):
    """Torch implementation of :class:`gunpowder.nodes.GenericTrain`.

    Args:

        model (subclass of ``torch.nn.Module``):

            The model to train.

        loss:

            The torch loss to use.

        optimizer:

            The torch optimizer to use.

        inputs (``dict``, ``string`` or ``int`` -> :class:`ArrayKey`):

            Dictionary from the names of input tensors (argument names of the
            ``forward`` method) in the model to array keys.

        loss_inputs (``dict``, ``string`` or ``int`` -> :class:`ArrayKey`):

            Dictionary with the names of input variables to the loss function as
            keys, and ArrayKeys containing the desired data as values. Keys can
            be either strings or integers. If the key is an integer, it will
            be treated as a positional argument to the loss function, a
            string will be used as a named argument

        outputs (``dict``, ``string`` or ``int`` -> :class:`ArrayKey`):

            Dictionary from the names of tensors in the network to array
            keys. If the key is a string, the tensor will be retrieved
            by checking the model for an attribute with they key as its name.
            If the key is an integer, it is interpreted as a tuple index of
            the outputs of the network.
            New arrays will be generated by this node for each entry (if
            requested downstream).

        gradients (``dict``, ``string`` or ``int`` -> :class:`ArrayKey`, optional):

            Dictionary from the names of tensors in the network to array
            keys. If the key is a string, the tensor will be retrieved
            by checking the model for an attribute with they key as its name.
            If the key is an integer, it is interpreted as a tuple index of
            the outputs of the network.
            Instead of the actual array, the gradient of the array with respect
            to the loss will be generated and saved.

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            Used to set the specs of generated arrays (at the moment only
            ``output``). This is useful to set the ``voxel_size``, for example,
            if they differ from the voxel size of the input arrays. Only fields
            that are not ``None`` in the given :class:`ArraySpec` will be used.

        checkpoint_basename (``string``, optional):

            The basename used for checkpoint files. Defaults to ``model``.

        save_every (``int``, optional):

            After how many iterations to create a checkpoint to store the
            learnt weights.

        log_dir (``string``, optional):

            Directory for saving tensorboard summaries.

        log_every (``int``, optional):

            After how many iterations to write out tensorboard summaries.

        spawn_subprocess (``bool``, optional):

            Whether to run the ``train_step`` in a separate process. Default is false.

        device (``str``, optional):

            Accepts a cuda gpu specifically to train on (e.g. `cuda:1`, `cuda:2`), helps in multi-card systems.
            defaults to ``cuda``

    """

    def __init__(
        self,
        model,
        loss,
        optimizer,
        inputs: Dict[Union[str, int], ArrayKey],
        outputs: Dict[Union[int, str], ArrayKey],
        loss_inputs: Dict[Union[int, str], ArrayKey],
        gradients: Dict[Union[int, str], ArrayKey] = {},
        array_specs: Optional[Dict[ArrayKey, ArraySpec]] = None,
        checkpoint_basename: str = "model",
        save_every: int = 2000,
        log_dir: Optional[str] = None,
        log_every: int = 1,
        spawn_subprocess: bool = False,
        device: str = "cuda",
    ):
        if not model.training:
            logger.warning(
                "Model is in evaluation mode during training. "
                "Consider using model.train()"
            )

        # not yet implemented
        gradients = gradients
        loss_inputs = {f"loss_{k}": v for k, v in loss_inputs.items()}
        all_inputs: dict[str | int, Any] = {
            f"{k}": v for k, v in inputs.items() if v not in outputs.values()
        }
        all_inputs.update(
            {k: v for k, v in loss_inputs.items() if v not in outputs.values()}
        )

        super(Train, self).__init__(
            all_inputs,
            outputs,
            gradients,
            array_specs,
            spawn_subprocess=spawn_subprocess,
        )

        self.model = model
        self.loss = loss
        self.optimizer = optimizer
        self.loss_inputs = loss_inputs
        self.checkpoint_basename = checkpoint_basename
        self.save_every = save_every
        self.dev = device

        self.iteration = 0

        if not isinstance(tensorboardX, NoSuchModule) and log_dir is not None:
            self.summary_writer = tensorboardX.SummaryWriter(log_dir)
            self.log_every = log_every
        else:
            self.summary_writer = None
            if log_dir is not None:
                logger.warning("log_dir given, but tensorboardX is not installed")

        self.intermediate_layers: dict[ArrayKey, Any] = {}
        self.register_hooks()

    def register_hooks(self):
        for key in self.outputs:
            if isinstance(key, str):
                layer = getattr(self.model, key)
                layer.register_forward_hook(self.create_hook(key))

    def create_hook(self, key):
        def save_layer(module, input, output):
            self.intermediate_layers[key] = output

        return save_layer

    def retain_gradients(self, request, outputs):
        for array_name, array_key in self.gradients.items():
            if array_key not in request:
                continue
            if isinstance(array_name, int):
                tensor = outputs[array_name]
            elif isinstance(array_name, str):
                tensor = getattr(self.model, array_name)
            else:
                raise RuntimeError(
                    "only ints and strings are supported as gradients keys"
                )
            tensor.retain_grad()

    def start(self):
        self.use_cuda = torch.cuda.is_available()
        # Issue: #188
        self.device = torch.device(self.dev if self.use_cuda else "cpu")

        try:
            self.model = self.model.to(self.device)
        except RuntimeError as e:
            raise RuntimeError(
                "Failed to move model to device. If you are using a child process "
                "to run your model, maybe you already initialized CUDA by sending "
                "your model to device in the main process."
            ) from e
        if isinstance(self.loss, torch.nn.Module):
            self.loss = self.loss.to(self.device)

        checkpoint, self.iteration = self._get_latest_checkpoint(
            self.checkpoint_basename
        )

        if checkpoint is not None:
            logger.info("Resuming training from iteration %d", self.iteration)
            logger.info("Loading %s", checkpoint)

            checkpoint = torch.load(checkpoint, map_location=self.device)
            self.model.load_state_dict(checkpoint["model_state_dict"])
            self.optimizer.load_state_dict(checkpoint["optimizer_state_dict"])

        else:
            logger.info("Starting training from scratch")

        logger.info("Using device %s", self.device)

    def train_step(self, batch, request):
        inputs = self.__collect_provided_inputs(batch)
        inputs = {k: torch.as_tensor(v, device=self.device) for k, v in inputs.items()}
        requested_outputs = self.__collect_requested_outputs(request)

        # keys are argument names of model forward pass
        device_input_args = []
        for i in range(len(inputs)):
            key = f"{i}"
            if key in inputs:
                device_input_args.append(inputs.pop(key))
            else:
                break
        device_input_kwargs = {k: v for k, v in inputs.items() if isinstance(k, str)}

        # get outputs. Keys are tuple indices or model attr names as in self.outputs
        self.optimizer.zero_grad()
        model_outputs = self.model(*device_input_args, **device_input_kwargs)
        if isinstance(model_outputs, tuple):
            outputs = {i: model_outputs[i] for i in range(len(model_outputs))}
        elif isinstance(model_outputs, torch.Tensor):
            outputs = {0: model_outputs}
        else:
            raise RuntimeError(
                "Torch train node only supports return types of tuple",
                f"and torch.Tensor from model.forward(). not {type(model_outputs)}",
            )
        outputs.update(self.intermediate_layers)

        # Some inputs to the loss should come from the batch, not the model
        provided_loss_inputs = self.__collect_provided_loss_inputs(batch)

        device_loss_inputs = {
            k: torch.as_tensor(v, device=self.device)
            for k, v in provided_loss_inputs.items()
        }

        # Some inputs to the loss function should come from the outputs of the model
        # Update device loss inputs with tensors from outputs if available
        flipped_outputs = {v: outputs[k] for k, v in self.outputs.items()}
        device_loss_inputs = {
            k[5:]: flipped_outputs.get(v, device_loss_inputs.get(k))
            for k, v in self.loss_inputs.items()
        }

        device_loss_args = []
        for i in range(len(device_loss_inputs)):
            key = f"{i}"
            if key in device_loss_inputs:
                device_loss_args.append(device_loss_inputs.pop(key))
            else:
                break
        device_loss_kwargs = {}
        for k, v in list(device_loss_inputs.items()):
            if isinstance(k, str):
                device_loss_kwargs[k] = device_loss_inputs.pop(k)
        assert (
            len(device_loss_inputs) == 0
        ), f"Not all loss inputs could be interpreted. Failed keys: {device_loss_inputs.keys()}"

        self.retain_gradients(request, outputs)

        logger.debug("model outputs: %s", {k: v.shape for k, v in outputs.items()})
        logger.debug(
            "loss inputs: %s %s",
            [v.shape for v in device_loss_args],
            {k: v.shape for k, v in device_loss_kwargs.items()},
        )
        loss = self.loss(*device_loss_args, **device_loss_kwargs)
        loss.backward()
        self.optimizer.step()

        # add requested model outputs to batch
        for array_key, array_name in requested_outputs.items():
            spec = self.spec[array_key].copy()
            spec.roi = request[array_key].roi
            batch.arrays[array_key] = Array(
                outputs[array_name].cpu().detach().numpy(), spec
            )

        for array_name, array_key in self.gradients.items():
            if array_key not in request:
                continue
            if isinstance(array_name, int):
                tensor = outputs[array_name]
            elif isinstance(array_name, str):
                tensor = getattr(self.model, array_name)
            else:
                raise RuntimeError(
                    "only ints and strings are supported as gradients keys"
                )
            spec = self.spec[array_key].copy()
            spec.roi = request[array_key].roi
            batch.arrays[array_key] = Array(tensor.grad.cpu().detach().numpy(), spec)

        batch.loss = loss.cpu().detach().numpy()
        self.iteration += 1
        batch.iteration = self.iteration

        if batch.iteration % self.save_every == 0:
            checkpoint_name = self._checkpoint_name(
                self.checkpoint_basename, batch.iteration
            )

            logger.info("Creating checkpoint %s", checkpoint_name)

            torch.save(
                {
                    "model_state_dict": self.model.state_dict(),
                    "optimizer_state_dict": self.optimizer.state_dict(),
                },
                checkpoint_name,
            )

        if self.summary_writer and batch.iteration % self.log_every == 0:
            self.summary_writer.add_scalar("loss", batch.loss, batch.iteration)

    def __collect_requested_outputs(self, request):
        array_outputs = {}

        for output_name, array_key in self.outputs.items():
            if array_key in request:
                array_outputs[array_key] = output_name

        return array_outputs

    def __collect_provided_inputs(self, batch):
        return self.__collect_provided_arrays(
            {
                k: v
                for k, v in self.inputs.items()
                if (isinstance(k, int) or k not in self.loss_inputs)
            },
            batch,
        )

    def __collect_provided_loss_inputs(self, batch):
        return self.__collect_provided_arrays(
            self.loss_inputs, batch, expect_missing_arrays=True
        )

    def __collect_provided_arrays(self, reference, batch, expect_missing_arrays=False):
        arrays = {}

        for array_name, array_key in reference.items():
            if isinstance(array_key, ArrayKey):
                msg = f"batch does not contain {array_key}, array {array_name} will not be set"
                if array_key in batch.arrays:
                    arrays[array_name] = batch.arrays[array_key].data
                elif not expect_missing_arrays:
                    logger.warn(msg)
                else:
                    logger.debug(msg)
            elif isinstance(array_key, np.ndarray):
                arrays[array_name] = array_key
            elif isinstance(array_key, str):
                arrays[array_name] = getattr(batch, array_key)
            else:
                raise Exception(
                    "Unknown network array key {}, can't be given to network".format(
                        array_key
                    )
                )

        return arrays



import logging
import multiprocessing as mp
from random import randrange

from gunpowder.array import Array
from gunpowder.array_spec import ArraySpec
from gunpowder.batch_request import BatchRequest
from gunpowder.coordinate import Coordinate

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class IterateLocations(BatchFilter):
    """Iterates over the nodes in a graph and centers
    batches at their locations. The iteration is thread safe.

    Args:
        graph (:class:`GraphKey`): Key of graph to read nodes from

        roi (:class:`Roi`): Roi within which to read and iterate over nodes.
            Defaults to None, which queries the whole Roi of the upstream graph
            source

        node_id (:class:`ArrayKey`, optional): Nonspatial array key in which to
            store the id of the "current" node in graph.  Default is None, in
            which case no attribute is stored and there is no way to tell which
            node is being considered.

        choose_randomly (bool): If true, choose nodes randomly with
            replacement. Default is false, which loops over the list.
    """

    __global_index = mp.Value("i", -1)
    visited_all = mp.Value("b", False)

    def __init__(self, graph, roi=None, node_id=None, choose_randomly=False):
        self.graph = graph
        self.roi = roi
        self.node_id = node_id
        self.choose_randomly = choose_randomly
        self.nodes = None
        self.coordinates = None
        self.local_index = None
        self.shift = None

    def setup(self):
        upstream = self.get_upstream_provider()
        self.upstream_spec = upstream.spec
        assert self.graph in self.upstream_spec, (
            "Upstream provider does not have graph %s" % self.graph
        )
        query_spec = self.upstream_spec.graph_specs[self.graph].copy()
        if self.roi:
            query_spec.roi = query_spec.roi.intersect(self.roi)
        # TODO: For scalability, scan upstream roi in blocks instead of
        #       storing all nodes in memory
        logger.info("Requesting all %s points in roi %s", self.graph, query_spec.roi)
        upstream_request = BatchRequest({self.graph: query_spec})
        upstream_batch = upstream.request_batch(upstream_request)
        self.nodes = list(upstream_batch[self.graph].nodes)
        self.coordinates = [node.location for node in self.nodes]
        assert (
            len(self.coordinates) > 0
        ), "Graph  %s doesn't have nodes to iterate over in roi %s" % (
            self.graph,
            self.roi,
        )

        # clear bounding boxes of all provided arrays and points
        for key, spec in self.spec.items():
            if spec.roi is not None:
                spec.roi.shape = Coordinate((None,) * spec.roi.dims)
                self.updates(key, spec)
        if self.node_id is not None:
            self.provides(self.node_id, ArraySpec(nonspatial=True))

    def prepare(self, request):
        logger.debug("request: %s", request.array_specs)
        logger.debug("my spec: %s", self.spec)

        lcm_voxel_size = self.spec.get_lcm_voxel_size(request.array_specs.keys())
        if lcm_voxel_size is None:
            ndims = len(self.coordinates[0])
            lcm_voxel_size = Coordinate((1,) * ndims)

        # shift to center
        total_roi = request.get_total_roi()
        request_center = total_roi.shape / 2 + total_roi.offset

        self.shift = self._get_next_shift(request_center, lcm_voxel_size)
        max_tries = 15
        tries = 0
        while not self.__check_shift(request):
            logger.warning(
                "Location %s (shift %s) skipped"
                % (self.coordinates[self.local_index], self.shift)
            )
            assert tries < max_tries, (
                "Unable to find valid shift after %d tries",
                tries,
            )
            self.shift = self._get_next_shift(request_center, lcm_voxel_size)
            tries += 1

        # Set shift for all requests
        for specs_type in [request.array_specs, request.graph_specs]:
            for key, spec in specs_type.items():
                if isinstance(spec, ArraySpec) and spec.nonspatial:
                    continue
                roi = spec.roi.shift(self.shift)
                specs_type[key].roi = roi

        logger.debug(
            "{}'th ({}) shift selected: {}".format(
                self.local_index, self.coordinates[self.local_index], self.shift
            )
        )

    def process(self, batch, request):
        if self.node_id:
            node_id = self.nodes[self.local_index].id
            spec = self.spec[self.node_id].copy()
            batch[self.node_id] = Array([node_id], spec)

        # reset ROIs to request
        for array_key, spec in request.array_specs.items():
            batch.arrays[array_key].spec.roi = spec.roi

        for graph_key, spec in request.graph_specs.items():
            batch.graphs[graph_key].spec.roi = spec.roi

        # change shift point locations to lie within roi
        for graph_key in request.graph_specs.keys():
            batch.graphs[graph_key].shift(-self.shift)

    def _get_next_shift(self, center_shift, voxel_size):
        # gets next coordinate from list
        if self.choose_randomly:
            self.local_index = randrange(len(self.coordinates))
        else:
            with IterateLocations.__global_index.get_lock():
                IterateLocations.__global_index.value += 1
                if IterateLocations.__global_index.value == len(self.coordinates) - 1:
                    logger.info("After this request, all points have been visited")
                    with IterateLocations.visited_all.get_lock():
                        IterateLocations.visited_all.value = True
                if IterateLocations.__global_index.value == len(self.coordinates):
                    logger.warning("Ran out of locations, looping list")
                self.local_index = IterateLocations.__global_index.value % len(
                    self.coordinates
                )
        next_shift = Coordinate(self.coordinates[self.local_index]) - center_shift

        logger.debug("Shift before rounding: %s" % str(next_shift))
        # make sure shift is a multiple of voxel size (round to nearest)
        next_shift = Coordinate(
            [
                int(vs * round(float(shift) / vs))
                for vs, shift in zip(voxel_size, next_shift)
            ]
        )
        logger.debug("Shift after rounding: %s" % str(next_shift))
        return next_shift

    def __check_shift(self, request):
        for key, spec in request.items():
            if isinstance(spec, ArraySpec) and spec.nonspatial:
                continue
            request_roi = spec.roi
            if key in self.upstream_spec:
                provided_roi = self.upstream_spec[key].roi
            else:
                raise Exception("Requested %s, but upstream does not provide it." % key)
            shifted_roi = request_roi.shift(self.shift)
            if not provided_roi.contains(shifted_roi):
                logger.warning(
                    ("Provided roi %s for key %s does notcontain shifted roi %s"),
                    provided_roi,
                    key,
                    shifted_roi,
                )
                return False
        return True



import logging
from typing import List

import numpy as np

from gunpowder.array import ArrayKey
from gunpowder.batch import Batch
from gunpowder.batch_request import BatchRequest

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class Squeeze(BatchFilter):
    """Squeeze a batch at a given axis

    Args:
        arrays (List[ArrayKey]): ArrayKeys to squeeze.
        axis: Position of the single-dimensional axis to remove, defaults to 0.
    """

    def __init__(self, arrays: List[ArrayKey], axis: int = 0):
        self.arrays = arrays
        self.axis = axis

    def setup(self):
        self.enable_autoskip()
        for array in self.arrays:
            self.updates(array, self.spec[array].copy())

    def prepare(self, request):
        deps = BatchRequest()
        for array in self.arrays:
            if array in request:
                deps[array] = request[array].copy()
        return deps

    def process(self, batch, request):
        outputs = Batch()
        for array in self.arrays:
            if array in batch:
                if not batch[array].spec.nonspatial:
                    spatial_dims = request[array].roi.dims
                    if self.axis >= batch[array].data.ndim - spatial_dims:
                        raise ValueError(
                            (
                                f"Squeeze.axis={self.axis} not permitted. "
                                "Squeeze only supported for "
                                "non-spatial dimensions of Array."
                            )
                        )

                outputs[array] = batch[array]
                outputs[array].data = np.squeeze(batch[array].data, self.axis)
                logger.debug(f"{array} shape: {outputs[array].data.shape}")

        return outputs



import copy
from typing import TYPE_CHECKING

from .batch_provider import BatchProvider

if TYPE_CHECKING:
    from gunpowder import Array, ArrayKey, Batch


class ArraySource(BatchProvider):
    def __init__(self, key: "ArrayKey", array: "Array"):
        self.key = key
        self.array = array

    def setup(self):
        self.provides(self.key, self.array.spec.copy())

    def provide(self, request):
        outputs = Batch()
        if self.array.spec.nonspatial:
            outputs[self.key] = copy.deepcopy(self.array)
        else:
            outputs[self.key] = copy.deepcopy(self.array.crop(request[self.key].roi))
        return outputs



from collections.abc import MutableMapping
from typing import Union

import zarr
from zarr._storage.store import BaseStore


class ZarrFile:
    """To be used as a context manager, similar to h5py.File."""

    def __init__(self, store: Union[BaseStore, MutableMapping, str], mode="a"):
        self.store = store
        self.mode = mode

    def __enter__(self):
        return zarr.open(self.store, mode=self.mode)

    def __exit__(self, *args):
        pass



import logging

logger = logging.getLogger(__name__)


class build(object):
    def __init__(self, pipeline):
        self.pipeline = pipeline

    def __enter__(self):
        try:
            self.pipeline.setup()
        except:
            logger.error(
                "something went wrong during the setup of the pipeline, calling tear down"
            )
            self.pipeline.internal_teardown()
            logger.debug("tear down completed")
            raise
        return self.pipeline

    def __exit__(self, type, value, traceback):
        logger.debug("leaving context, tearing down pipeline")
        self.pipeline.internal_teardown()
        logger.debug("tear down completed")



import logging

from gunpowder.array import Array, ArrayKey
from gunpowder.batch import Batch

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class AsType(BatchFilter):
    """Cast arrays to a different datatype (ex: np.float32 --> np.uint8).

    Args:

        source (:class:`ArrayKey`):

            The key of the array to cast.

        target_dtype (str or dtype):

            The voxel size of the target.

        target (:class:`ArrayKey`, optional):

            The key of the array to store the cast ``source``.

    """

    def __init__(self, source, target_dtype, target=None):
        assert isinstance(source, ArrayKey)
        if target is not None:
            assert isinstance(target, ArrayKey)
            self.target = target
        else:
            self.target = source

        self.source = source
        self.target_dtype = target_dtype

    def setup(self):
        spec = self.spec[self.source].copy()
        spec.dtype = self.target_dtype
        if self.target is not self.source:
            self.provides(self.target, spec)
        else:
            self.updates(self.source, spec)
        self.enable_autoskip()

    def process(self, batch, request):
        source = batch.arrays[self.source]
        source_data = source.data

        cast_data = source_data.astype(self.target_dtype)

        target_spec = source.spec.copy()
        target_spec.dtype = cast_data.dtype
        target_array = Array(cast_data, target_spec)

        # create output array
        outputs = Batch()
        outputs.arrays[self.target] = target_array

        return outputs



import logging
import os

import numpy as np

from gunpowder.batch_request import BatchRequest
from gunpowder.ext import ZarrFile, h5py

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class Snapshot(BatchFilter):
    """Save a passing batch in an HDF or Zarr file.

    The default behaviour is to periodically save a snapshot after
    ``every`` iterations.

    Data-dependent criteria for saving can be implemented by subclassing and
    overwriting :func:`write_if`. This method is applied as an additional
    filter to the batches picked for periodic saving. It should return ``True``
    if a batch meets the criteria for saving.

    Args:

        dataset_names (``dict``, :class:`ArrayKey` -> ``string``):

            A dictionary from array keys to names of the datasets to store them
            in.

        output_dir (``string``):

            The directory to save the snapshots. Will be created, if it does
            not exist.

        output_filename (``string``):

            Template for output filenames. ``{id}`` in the string will be
            replaced with the ID of the batch. ``{iteration}`` with the training
            iteration (if training was performed on this batch). Snapshot will
            be saved as zarr file if output_filename ends in ``.zarr`` and as
            HDF otherwise.

        every (``int``):

            How often to save a batch. ``every=1`` indicates that every batch
            will be stored, ``every=2`` every second and so on. By default,
            every batch will be stored.

        additional_request (:class:`BatchRequest`):

            An additional batch request to merge with the passing request, if a
            snapshot is to be made. If not given, only the arrays that are in
            the batch anyway are recorded. This is useful to request additional
            arrays like loss gradients for visualization that are otherwise not
            needed.

        compression_type (``string`` or ``int``):

            Compression strategy.  Legal values are ``gzip``, ``szip``,
            ``lzf``. If an integer between 1 and 10, this indicates ``gzip``
            compression level.

        dataset_dtypes (``dict``, :class:`ArrayKey` -> data type):

            A dictionary from array keys to datatype (eg. ``np.int8``). If
            given, arrays are stored using this type. The original arrays
            within the pipeline remain unchanged.

        store_value_range (``bool``):

            If set to ``True``, store range of values in data set attributes.
    """

    def __init__(
        self,
        dataset_names,
        output_dir="snapshots",
        output_filename="{iteration}.zarr",
        every=1,
        additional_request=None,
        compression_type=None,
        dataset_dtypes=None,
        store_value_range=False,
    ):
        self.dataset_names = dataset_names
        self.output_dir = output_dir
        self.output_filename = output_filename
        self.every = max(1, every)
        self.additional_request = (
            BatchRequest() if additional_request is None else additional_request
        )
        self.n = 0
        self.compression_type = compression_type
        self.store_value_range = store_value_range
        if dataset_dtypes is None:
            self.dataset_dtypes = {}
        else:
            self.dataset_dtypes = dataset_dtypes

        self.mode = "w"
        self.id = 0

    def write_if(self, batch):
        """To be implemented in subclasses.

        This function is run in :func:`process` and acts as a data-dependent
        filter for saving snapshots.

        Args:

            batch (:class:`Batch`):

                The batch received from upstream.

        Returns:

            ``True`` if ``batch`` should be written to snapshot, ``False``
            otherwise.
        """

        return True

    def setup(self):
        for key, _ in self.additional_request.items():
            assert key in self.dataset_names, (
                "%s requested but not in dataset_names" % key
            )

        for array_key in self.additional_request.array_specs.keys():
            spec = self.spec[array_key]
            self.updates(array_key, spec)
        for graph_key in self.additional_request.graph_specs.keys():
            spec = self.spec[graph_key]
            self.updates(graph_key, spec)

    def prepare(self, request):
        deps = BatchRequest()
        for key, spec in request.items():
            if key in self.dataset_names:
                deps[key] = spec

        self.record_snapshot = self.n % self.every == 0

        if self.record_snapshot:
            # append additional array requests, don't overwrite existing ones
            for array_key, spec in self.additional_request.array_specs.items():
                if array_key not in deps:
                    deps[array_key] = spec
            for graph_key, spec in self.additional_request.graph_specs.items():
                if graph_key not in deps:
                    deps[graph_key] = spec

            for key in self.dataset_names.keys():
                assert key in deps, "%s wanted for %s, but not in request." % (
                    key,
                    self.name(),
                )

        return deps

    def process(self, batch, request):
        self.id += 1
        if self.record_snapshot and self.write_if(batch):
            try:
                os.makedirs(self.output_dir)
            except:
                pass

            snapshot_name = os.path.join(
                self.output_dir,
                self.output_filename.format(
                    id=str(self.id).zfill(8), iteration=int(batch.iteration or self.id)
                ),
            )
            logger.info("saving to %s" % snapshot_name)
            if snapshot_name.endswith(".hdf"):
                open_func = h5py.File
            elif snapshot_name.endswith(".zarr"):
                open_func = ZarrFile
            else:
                logger.warning("ambiguous file type")
                open_func = h5py.File

            with open_func(snapshot_name, self.mode) as f:
                for array_key, array in batch.arrays.items():
                    if array_key not in self.dataset_names:
                        continue

                    ds_name = self.dataset_names[array_key]

                    if array_key in self.dataset_dtypes:
                        dtype = self.dataset_dtypes[array_key]
                        dataset = f.create_dataset(
                            name=ds_name,
                            data=array.data.astype(dtype),
                            compression=self.compression_type,
                        )

                    else:
                        dataset = f.create_dataset(
                            name=ds_name,
                            data=array.data,
                            compression=self.compression_type,
                        )

                    if not array.spec.nonspatial:
                        if array.spec.roi is not None:
                            dataset.attrs["offset"] = array.spec.roi.offset
                        dataset.attrs["voxel_size"] = self.spec[array_key].voxel_size

                    if self.store_value_range:
                        dataset.attrs["value_range"] = (
                            array.data.min().item(),
                            array.data.max().item(),
                        )

                    # if array has attributes, add them to the dataset
                    for attribute_name, attribute in array.attrs.items():
                        dataset.attrs[attribute_name] = attribute

                for graph_key, graph in batch.graphs.items():
                    if graph_key not in self.dataset_names:
                        continue

                    ds_name = self.dataset_names[graph_key]

                    node_ids = []
                    locations = []
                    edges = []
                    for node in graph.nodes:
                        node_ids.append(node.id)
                        locations.append(node.location)
                    for edge in graph.edges:
                        edges.append((edge.u, edge.v))

                    f.create_dataset(
                        name=f"{ds_name}-ids",
                        data=np.array(node_ids, dtype=int),
                        compression=self.compression_type,
                    )
                    f.create_dataset(
                        name=f"{ds_name}-locations",
                        data=np.array(locations),
                        compression=self.compression_type,
                    )
                    f.create_dataset(
                        name=f"{ds_name}-edges",
                        data=np.array(edges),
                        compression=self.compression_type,
                    )

                if batch.loss is not None:
                    f["/"].attrs["loss"] = float(batch.loss)

        self.n += 1



import numpy as np

from gunpowder.array import Array
from gunpowder.array_spec import ArraySpec

from .batch_provider import BatchProvider


class RandomProvider(BatchProvider):
    """Randomly selects one of the upstream providers::

        (a, b, c) + RandomProvider()

    will create a provider that randomly relays requests to providers ``a``,
    ``b``, or ``c``. Array and point keys of ``a``, ``b``, and ``c`` should be
    the same.

    Args:
        probabilities (1-D array-like, optional):

            An optional list of
            probabilities for choosing upstream providers, given in the
            same order. Probabilities do not need to be normalized. Default
            is ``None``, corresponding to equal probabilities.

        random_provider_key (``ArrayKey``):

            If provided, this node will store the index of the chosen random
            provider in a nonspatial array.
    """

    def __init__(self, probabilities=None, random_provider_key=None):
        self.probabilities = probabilities
        self.random_provider_key = random_provider_key

        # automatically normalize probabilities to sum to 1
        if self.probabilities is not None:
            self.probabilities = [
                float(x) / np.sum(probabilities) for x in self.probabilities
            ]

    def setup(self):
        self.enable_placeholders()
        assert (
            len(self.get_upstream_providers()) > 0
        ), "at least one batch provider must be added to the RandomProvider"
        if self.probabilities is not None:
            assert len(self.get_upstream_providers()) == len(self.probabilities), (
                "if probabilities are specified, they "
                "need to be given for each batch "
                "provider added to the RandomProvider"
            )

        common_spec = None

        # advertise outputs only if all upstream providers have them
        for provider in self.get_upstream_providers():
            if common_spec is None:
                common_spec = provider.spec.copy()
            else:
                for key, spec in list(common_spec.items()):
                    if key not in provider.spec:
                        del common_spec[key]

        for key, spec in common_spec.items():
            self.provides(key, spec)

        if self.random_provider_key is not None:
            self.provides(self.random_provider_key, ArraySpec(nonspatial=True))

    def provide(self, request):
        if self.random_provider_key is not None:
            del request[self.random_provider_key]

        i = np.random.choice(
            range(len(self.get_upstream_providers())), p=self.probabilities
        )
        provider = self.get_upstream_providers()[i]
        batch = provider.request_batch(request)
        if self.random_provider_key is not None:
            batch[self.random_provider_key] = Array(
                np.array(i), ArraySpec(nonspatial=True)
            )
        return batch



import copy
import time
from warnings import warn

from .array import ArrayKey
from .array_spec import ArraySpec
from .graph import GraphKey
from .graph_spec import GraphSpec
from .provider_spec import ProviderSpec
from .roi import Roi


class BatchRequest(ProviderSpec):
    """A collection of (possibly partial) :class:`ArraySpec` and
    :class:`GraphSpec` forming a request.

    Inherits from :class:`ProviderSpec`.

    Additional Kwargs:

        random_seed (``int``):

            The random seed that will be associated with this batch to
            guarantee deterministic and repeatable batch requests.

    """

    def __init__(self, *args, random_seed=None, **kwargs):
        self._random_seed = random_seed
        super().__init__(*args, **kwargs)

    def add(self, key, shape, voxel_size=None, directed=None, placeholder=False):
        """Convenience method to add an array or graph spec by providing only
        the shape of a ROI (in world units).

        A ROI with zero-offset will be generated. If more than one request is
        added, the ROIs with smaller shapes will be shifted to be centered in
        the largest one.

        Args:

            key (:class:`ArrayKey` or :class:`GraphKey`):

                The key for which to add a spec.

            shape (:class:`Coordinate`):

                A tuple containing the shape of the desired roi

            voxel_size (:class:`Coordinate`):

                A tuple contening the voxel sizes for each corresponding
                dimension
        """

        if isinstance(key, ArrayKey):
            spec = ArraySpec(placeholder=placeholder)
        elif isinstance(key, GraphKey):
            spec = GraphSpec(placeholder=placeholder, directed=directed)
        else:
            raise RuntimeError("Only ArrayKey or GraphKey can be added.")

        spec.roi = Roi((0,) * len(shape), shape)

        if voxel_size is not None:
            spec.voxel_size = voxel_size

        self[key] = spec
        self.__center_rois()

    def copy(self):
        """Create a copy of this request."""
        request_copy = copy.deepcopy(self)
        request_copy._update_random_seed()
        return request_copy

    def is_deterministic(self):
        """Return true if a random seed has been set for this request."""
        return self._random_seed is not None

    @property
    def random_seed(self):
        if not self.is_deterministic():
            return int(time.time() * 1e6) % (2**32)
        else:
            return self._random_seed

    def _update_random_seed(self):
        if not self.is_deterministic():
            return
        self._random_seed = hash((self._random_seed + 1) ** 2) % (2**32)

    def __center_rois(self):
        """Ensure that all ROIs are centered around the same location."""

        total_roi = self.get_total_roi()
        if total_roi is None:
            return

        center = total_roi.center

        for specs_type in [self.array_specs, self.graph_specs]:
            for key in specs_type:
                roi = specs_type[key].roi
                specs_type[key].roi = roi.shift(center - roi.center)

    def update_with(self, request):
        """Update current request with another"""

        assert isinstance(request, BatchRequest)

        merged = self.copy()

        for key, spec in request.items():
            if key not in merged:
                merged[key] = spec
            else:
                merged[key].update_with(spec)

        return merged

    def merge(self, request):
        """Merge another request with current request"""
        warn(
            "merge is deprecated! please use update_with "
            "as it accounts for spec metadata"
        )
        assert isinstance(request, BatchRequest)

        merged = self.copy()

        for key, spec in request.items():
            if key not in merged:
                merged[key] = spec
            else:
                if isinstance(spec, ArraySpec) and merged[key].nonspatial:
                    merged[key] = spec
                else:
                    merged[key].roi = merged[key].roi.union(spec.roi)

        return merged

    def __eq__(self, other):
        """
        Override equality check to allow batche requests with different
        seeds to still be checked. Otherwise equality check should
        never succeed.
        """

        if isinstance(other, self.__class__):
            other_dict = copy.deepcopy(other.__dict__)
            self_dict = copy.deepcopy(self.__dict__)
            other_dict.pop("_random_seed")
            self_dict.pop("_random_seed")
            return self_dict == other_dict
        return NotImplemented



import copy
import time

import numpy as np

from .freezable import Freezable


class Timing(Freezable):
    def __init__(self, node, method_name=None):
        self.__name = type(node).__name__
        self.__method_name = method_name
        self.__start = 0
        self.__first_start = 0
        self.__last_stop = 0
        self.__time = 0
        self.freeze()

    def start(self):
        self.__start = time.time()
        if self.__first_start == 0:
            self.__first_start = self.__start

    def stop(self):
        if self.__start == 0:
            return
        t = time.time()
        self.__time += t - self.__start
        self.__start = 0
        self.__last_stop = t

    def elapsed(self):
        """Accumulated time elapsed between calls to start() and stop()."""

        if self.__start == 0:
            return self.__time

        return self.__time + (time.time() - self.__start)

    def span(self):
        """Timestamps of the first call to start() and last call to stop()."""
        return self.__first_start, self.__last_stop

    def get_node_name(self):
        return self.__name

    def get_method_name(self):
        return self.__method_name


class TimingSummary(Freezable):
    """Holds repeated Timings of the same node/method to be queried for statistics."""

    def __init__(self):
        self.timings = []
        self.times = []
        self.freeze()

    def add(self, timing):
        """Add a Timing to this summary."""
        self.timings.append(timing)
        self.times.append(timing.elapsed())

    def merge(self, other):
        """Merge another summary into this one."""
        for timing in other.timings:
            self.add(timing)

    def counts(self):
        return len(self.times)

    def min(self):
        return np.min(self.times)

    def max(self):
        return np.max(self.times)

    def mean(self):
        return np.mean(self.times)

    def median(self):
        return np.median(self.times)


class ProfilingStats(Freezable):
    def __init__(self):
        self.__summaries = {}
        self.freeze()

    def add(self, timing):
        """Add a Timing instance. Timings are grouped by their class and method names."""

        node_name = timing.get_node_name()
        method_name = timing.get_method_name()
        id = (node_name, method_name)

        if id not in self.__summaries:
            self.__summaries[id] = TimingSummary()
        self.__summaries[id].add(copy.deepcopy(timing))

    def merge_with(self, other):
        """Combine statitics of two ProfilingStats."""

        for id, summary in other.__summaries.items():
            if id in self.__summaries:
                self.__summaries[id].merge(copy.deepcopy(summary))
            else:
                self.__summaries[id] = copy.deepcopy(summary)

    def get_timing_summaries(self):
        """Get a dictionary (node_name,method_name) -> TimingSummary."""
        return self.__summaries

    def get_timing_summary(self, node_name, method_name=None):
        """Get a :class:`TimingSummary` for the given node and method name."""

        if (node_name, method_name) not in self.__summaries:
            raise RuntimeError(
                "No timing summary for node %s, method %s" % (node_name, method_name)
            )

        return self.__summaries[(node_name, method_name)]

    def span(self):
        """Timestamps of the first call to start() and last call to stop() over
        all Timings added."""

        spans = [
            t.span()
            for (_, summary) in self.__summaries.items()
            for t in summary.timings
        ]
        first_start = min([span[0] for span in spans])
        last_stop = max([span[1] for span in spans])

        return first_start, last_stop

    def span_time(self):
        """Time between the first call to start() and last call to stop() over
        any timing."""

        start, stop = self.span()
        return stop - start



import glob
import logging
import multiprocessing
import re
import time

from gunpowder.array import ArrayKey
from gunpowder.array_spec import ArraySpec
from gunpowder.batch_request import BatchRequest
from gunpowder.nodes.batch_filter import BatchFilter
from gunpowder.producer_pool import NoResult, ProducerPool, WorkersDied

logger = logging.getLogger(__name__)


class TrainProcessDied(Exception):
    pass


class GenericTrain(BatchFilter):
    """Generic train node to perform one training iteration for each batch that
    passes through. This node alone does nothing and should be subclassed for
    concrete implementations.

    Args:

        inputs (dict): Dictionary from names of input layers in the network to
            :class:``ArrayKey`` or batch attribute name as string.

        outputs (dict): Dictionary from the names of output layers in the
            network to :class:``ArrayKey``. New arrays will be generated by
            this node for each entry (if requested downstream).

        gradients (dict): Dictionary from the names of output layers in the
            network to :class:``ArrayKey``. New arrays containing the
            gradient of an output with respect to the loss will be generated by
            this node for each entry (if requested downstream).

        array_specs (dict, optional): An optional dictionary of
            :class:`ArrayKey` to :class:`ArraySpec` to set the array specs
            generated arrays (``outputs`` and ``gradients``). This is useful
            to set the ``voxel_size``, for example, if they differ from the
            voxel size of the input arrays. Only fields that are not ``None``
            in the given :class:`ArraySpec` will be used.

        spawn_subprocess (bool, optional): Whether to run the ``train_step`` in
            a separate process. Default is false.
    """

    def __init__(
        self, inputs, outputs, gradients, array_specs=None, spawn_subprocess=False
    ):
        self.initialized = False

        self.inputs = inputs
        self.outputs = outputs
        self.gradients = gradients
        self.array_specs = {} if array_specs is None else array_specs
        self.spawn_subprocess = spawn_subprocess

        self.provided_arrays = list(self.outputs.values()) + list(
            self.gradients.values()
        )

    def setup(self):
        # get common voxel size of inputs, or None if they differ
        common_voxel_size = None
        for key in self.inputs.values():
            if not isinstance(key, ArrayKey):
                continue
            if self.spec[key].nonspatial:
                continue

            voxel_size = self.spec[key].voxel_size

            if common_voxel_size is None:
                common_voxel_size = voxel_size
            elif common_voxel_size != voxel_size:
                common_voxel_size = None
                break

        # announce provided outputs
        for key in self.provided_arrays:
            if key in self.array_specs:
                spec = self.array_specs[key].copy()
            else:
                spec = ArraySpec()

            if spec.voxel_size is None and not spec.nonspatial:
                assert common_voxel_size is not None, (
                    "There is no common voxel size of the inputs, and no "
                    "ArraySpec has been given for %s that defines "
                    "voxel_size." % key
                )

                spec.voxel_size = common_voxel_size

            if spec.interpolatable is None:
                # default for predictions
                spec.interpolatable = False

            self.provides(key, spec)

        if self.spawn_subprocess:
            # start training as a producer pool, so that we can gracefully exit if
            # anything goes wrong
            self.worker = ProducerPool([self._produce_train_batch], queue_size=1)
            self.batch_in = multiprocessing.Queue(maxsize=1)
            self.worker.start()
        else:
            self.start()
            self.initialized = True

    def prepare(self, request):
        deps = BatchRequest()
        for key in self.inputs.values():
            deps[key] = request[key]
        return deps

    def teardown(self):
        if self.spawn_subprocess:
            # signal "stop"
            self.batch_in.put((None, None))
            try:
                self.worker.get(timeout=2)
            except NoResult:
                pass
            self.worker.stop()
        else:
            self.stop()

    def process(self, batch, request):
        start = time.time()

        if self.spawn_subprocess:
            self.batch_in.put((batch, request))

            try:
                out = self.worker.get()
            except WorkersDied:
                raise TrainProcessDied()

            for array_key in self.provided_arrays:
                if array_key in request:
                    batch.arrays[array_key] = out.arrays[array_key]

            batch.loss = out.loss
            batch.iteration = out.iteration

        else:
            self.train_step(batch, request)

        time_of_iteration = time.time() - start

        logger.info(
            "Train process: iteration=%d loss=%f time=%f",
            batch.iteration,
            batch.loss,
            time_of_iteration,
        )

    def start(self):
        """To be implemented in subclasses.

        This method will be called before the first call to :fun:`train_step`,
        from the same process that :fun:`train_step` will be called from. Use
        this to initialize you solver and training hardware.
        """
        pass

    def train_step(self, batch, request):
        """To be implemented in subclasses.

        In this method, an implementation should perform one training iteration
        on the given batch. ``batch.loss`` and ``batch.iteration`` should be
        set. Output arrays should be created according to the given request
        and added to ``batch``."""
        raise NotImplementedError(
            "Class %s does not implement 'train_step'" % self.name()
        )

    def stop(self):
        """To be implemented in subclasses.

        This method will be called after the last call to :fun:`train_step`,
        from the same process that :fun:`train_step` will be called from. Use
        this to tear down you solver and free training hardware.
        """
        pass

    def _checkpoint_name(self, basename, iteration):
        return basename + "_checkpoint_" + "%i" % iteration

    def _get_latest_checkpoint(self, basename):
        def atoi(text):
            return int(text) if text.isdigit() else text

        def natural_keys(text):
            return [atoi(c) for c in re.split(r"(\d+)", text)]

        checkpoints = glob.glob(basename + "_checkpoint_*")
        checkpoints.sort(key=natural_keys)

        if len(checkpoints) > 0:
            checkpoint = checkpoints[-1]
            iteration = int(checkpoint.split("_")[-1])
            return checkpoint, iteration

        return None, 0

    def _produce_train_batch(self):
        """Process one train batch."""

        if not self.initialized:
            self.start()
            self.initialized = True

        batch, request = self.batch_in.get()

        # stop signal
        if batch is None:
            self.stop()
            return None

        self.train_step(batch, request)

        return batch



from __future__ import absolute_import

from .add_blobs_from_points import AddBlobsFromPoints
from .add_boundary_distance_gradients import AddBoundaryDistanceGradients
from .add_gt_mask_exclusive_zone import AddGtMaskExclusiveZone
from .add_nonsymmetric_affinities import AddNonsymmetricAffinities
from .add_vector_map import AddVectorMap
from .dvid_partner_annotation_source import DvidPartnerAnnotationSource
from .hdf5_points_source import Hdf5PointsSource
from .prepare_malis import PrepareMalis
from .zero_out_const_sections import ZeroOutConstSections



__major__ = 1
__minor__ = 4
__patch__ = 1
__tag__ = ""
__version__ = "{}.{}.{}{}".format(__major__, __minor__, __patch__, __tag__).strip(".")


class _Version(object):
    def major(self):
        return __major__

    def minor(self):
        return __minor__

    def patch(self):
        return __patch__

    def tag(self):
        return __tag__

    def version(self):
        return __version__

    def __str__(self):
        return self.version()


_version = _Version()

if __name__ == "__main__":
    print(_version)



import copy
from typing import TYPE_CHECKING

from ..batch import Batch
from .batch_provider import BatchProvider

if TYPE_CHECKING:
    from gunpowder import Graph, GraphKey


class GraphSource(BatchProvider):
    def __init__(self, key: "GraphKey", graph: "Graph"):
        self.key = key
        self.graph = graph

    def setup(self):
        self.provides(self.key, self.graph.spec)

    def provide(self, request):
        outputs = Batch()
        outputs[self.key] = copy.deepcopy(
            self.graph.crop(request[self.key].roi).trim(request[self.key].roi)
        )
        return outputs



import logging
from collections import deque

from gunpowder.producer_pool import ProducerPool
from gunpowder.profiling import Timing

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class WorkersDiedException(Exception):
    pass


class PreCache(BatchFilter):
    """Pre-cache repeated equal batch requests. For the first of a series of
    equal batch request, a set of workers is spawned to pre-cache the batches
    in parallel processes. This way, subsequent requests can be served quickly.

    A note on changing the requests sent to `PreCache`.
    Given requests A and B, if requests are sent in the sequence:
    A, ..., A, B, A, ..., A, B, A, ...
    Precache will build a Queue of batches that satisfy A, and handle requests
    B on demand. This prevents `PreCache` from discarding the queue on every
    SnapshotRequest.
    However if B request replace A as the most common request, i.e.:
    A, A, A, ..., A, B, B, B, ...,
    `PreCache` will discard the A queue and build a B queue after it has seen
    more B requests than A requests out of the last 5 requests.

    This node only makes sense if:

    1. Incoming batch requests are repeatedly the same.
    2. There is a source of randomness in upstream nodes.

    Args:

        cache_size (``int``):

            How many batches to hold at most in the cache.

        num_workers (``int``):

            How many processes to spawn to fill the cache.
    """

    def __init__(self, cache_size=50, num_workers=20):
        self.current_request = None
        self.workers = None
        self.cache_size = cache_size
        self.num_workers = num_workers

        # keep track of recent requests
        self.last_5 = deque(
            [
                None,
            ]
            * 5,
            maxlen=5,
        )

    def teardown(self):
        if self.workers is not None:
            self.workers.stop()

    def provide(self, request):
        timing = Timing(self)
        timing.start()

        # update recent requests
        self.last_5.popleft()
        self.last_5.append(request)

        if request != self.current_request:
            current_count = sum(
                [
                    recent_request == self.current_request
                    for recent_request in self.last_5
                ]
            )
            new_count = sum(
                [recent_request == request for recent_request in self.last_5]
            )
            if new_count > current_count or self.current_request is None:
                if self.workers is not None:
                    logger.info("new request received, stopping current workers...")
                    self.workers.stop()

                self.current_request = request.copy()

                logger.info(
                    "starting new set of workers (%s, cache size %s)...",
                    self.num_workers,
                    self.cache_size,
                )
                self.workers = ProducerPool(
                    [self._run_worker for _ in range(self.num_workers)],
                    queue_size=self.cache_size,
                )
                self.workers.start()

                logger.debug("getting batch from queue...")
                batch = self.workers.get()

                timing.stop()
                batch.profiling_stats.add(timing)

            else:
                logger.debug("Resolving new request sequentially")
                batch = self.get_upstream_provider().request_batch(request)

                timing.stop()
                batch.profiling_stats.add(timing)

        else:
            logger.debug("getting batch from queue...")
            batch = self.workers.get()

            timing.stop()
            batch.profiling_stats.add(timing)

        return batch

    def _run_worker(self):
        request = self.current_request.copy()
        return self.get_upstream_provider().request_batch(request)



import tensorflow as tf


def conv_pass(
    fmaps_in,
    kernel_size,
    num_fmaps,
    num_repetitions,
    activation="relu",
    name="conv_pass",
):
    """Create a convolution pass::

        f_in --> f_1 --> ... --> f_n

    where each ``-->`` is a convolution followed by a (non-linear) activation
    function and ``n`` ``num_repetitions``. Each convolution will decrease the
    size of the feature maps by ``kernel_size-1``.

    Args:

        f_in:

            The input tensor of shape ``(batch_size, channels, depth, height,
            width)`` or ``(batch_size, channels, height, width)``.

        kernel_size:

            Size of the kernel. Forwarded to the tensorflow convolution layer.

        num_fmaps:

            The number of feature maps to produce with each convolution.

        num_repetitions:

            How many convolutions to apply.

        activation:

            Which activation to use after a convolution. Accepts the name of any
            tensorflow activation function (e.g., ``relu`` for ``tf.nn.relu``).

    """

    fmaps = fmaps_in
    if activation is not None:
        activation = getattr(tf.nn, activation)

    conv_layer = getattr(
        tf.layers, {2: "conv2d", 3: "conv3d"}[fmaps_in.get_shape().ndims - 2]
    )

    for i in range(num_repetitions):
        fmaps = conv_layer(
            inputs=fmaps,
            filters=num_fmaps,
            kernel_size=kernel_size,
            padding="valid",
            data_format="channels_first",
            activation=activation,
            name=name + "_%i" % i,
        )

    return fmaps


def downsample(fmaps_in, factors, name="down"):
    pooling_layer = getattr(
        tf.layers,
        {2: "max_pooling2d", 3: "max_pooling3d"}[fmaps_in.get_shape().ndims - 2],
    )

    fmaps = pooling_layer(
        fmaps_in,
        pool_size=factors,
        strides=factors,
        padding="valid",
        data_format="channels_first",
        name=name,
    )

    return fmaps


def upsample(fmaps_in, factors, num_fmaps, activation="relu", name="up"):
    if activation is not None:
        activation = getattr(tf.nn, activation)

    conv_trans_layer = getattr(
        tf.layers,
        {2: "conv2d_transpose", 3: "conv3d_transpose"}[fmaps_in.get_shape().ndims - 2],
    )

    fmaps = conv_trans_layer(
        fmaps_in,
        filters=num_fmaps,
        kernel_size=factors,
        strides=factors,
        padding="valid",
        data_format="channels_first",
        activation=activation,
        name=name,
    )

    return fmaps


def crop_spatial(fmaps_in, shape):
    """Crop only the spacial dimensions to match shape.

    Args:

        fmaps_in:

            The input tensor.

        shape:

            A list (not a tensor) with the requested shape [_, _, z, y, x] or
            [_, _, y, x].
    """

    in_shape = fmaps_in.get_shape().as_list()

    offset = [0, 0] + [(in_shape[i] - shape[i]) // 2 for i in range(2, len(shape))]
    size = in_shape[0:2] + shape[2:]

    fmaps = tf.slice(fmaps_in, offset, size)

    return fmaps


def unet(
    fmaps_in, num_fmaps, fmap_inc_factor, downsample_factors, activation="relu", layer=0
):
    """Create a 2D or 3D U-Net::

        f_in --> f_left --------------------------->> f_right--> f_out
                    |                                   ^
                    v                                   |
                 g_in --> g_left ------->> g_right --> g_out
                             |               ^
                             v               |
                                   ...

    where each ``-->`` is a convolution pass (see ``conv_pass``), each `-->>` a
    crop, and down and up arrows are max-pooling and transposed convolutions,
    respectively.

    The U-Net expects tensors to have shape ``(batch=1, channels, depth, height,
    width)`` for 3D or ``(batch=1, channels, height, width)`` for 2D.

    This U-Net performs only "valid" convolutions, i.e., sizes of the feature
    maps decrease after each convolution.

    Args:

        fmaps_in:

            The input tensor.

        num_fmaps:

            The number of feature maps in the first layer. This is also the
            number of output feature maps.

        fmap_inc_factor:

            By how much to multiply the number of feature maps between layers.
            If layer 0 has ``k`` feature maps, layer ``l`` will have
            ``k*fmap_inc_factor**l``.

        downsample_factors:

            List of lists ``[z, y, x]`` or ``[y, x]`` to use to down- and
            up-sample the feature maps between layers.

        activation:

            Which activation to use after a convolution. Accepts the name of any
            tensorflow activation function (e.g., ``relu`` for ``tf.nn.relu``).

        layer:

            Used internally to build the U-Net recursively.
    """

    prefix = "    " * layer
    print(prefix + "Creating U-Net layer %i" % layer)
    print(prefix + "f_in: " + str(fmaps_in.shape))

    # convolve
    f_left = conv_pass(
        fmaps_in,
        kernel_size=3,
        num_fmaps=num_fmaps,
        num_repetitions=2,
        activation=activation,
        name="unet_layer_%i_left" % layer,
    )

    # last layer does not recurse
    bottom_layer = layer == len(downsample_factors)
    if bottom_layer:
        print(prefix + "bottom layer")
        print(prefix + "f_out: " + str(f_left.shape))
        return f_left

    # downsample
    g_in = downsample(
        f_left, downsample_factors[layer], "unet_down_%i_to_%i" % (layer, layer + 1)
    )

    # recursive U-net
    g_out = unet(
        g_in,
        num_fmaps=num_fmaps * fmap_inc_factor,
        fmap_inc_factor=fmap_inc_factor,
        downsample_factors=downsample_factors,
        activation=activation,
        layer=layer + 1,
    )

    print(prefix + "g_out: " + str(g_out.shape))

    # upsample
    g_out_upsampled = upsample(
        g_out,
        downsample_factors[layer],
        num_fmaps,
        activation=activation,
        name="unet_up_%i_to_%i" % (layer + 1, layer),
    )

    print(prefix + "g_out_upsampled: " + str(g_out_upsampled.shape))

    # copy-crop
    f_left_cropped = crop_spatial(f_left, g_out_upsampled.get_shape().as_list())

    print(prefix + "f_left_cropped: " + str(f_left_cropped.shape))

    # concatenate along channel dimension
    f_right = tf.concat([f_left_cropped, g_out_upsampled], 1)

    print(prefix + "f_right: " + str(f_right.shape))

    # convolve
    f_out = conv_pass(
        f_right,
        kernel_size=3,
        num_fmaps=num_fmaps,
        num_repetitions=2,
        name="unet_layer_%i_right" % layer,
    )

    print(prefix + "f_out: " + str(f_out.shape))

    return f_out



import logging

import numpy as np

from gunpowder.array import Array
from gunpowder.nodes.batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class AddNonsymmetricAffinities(BatchFilter):
    def __init__(
        self,
        affinity_vectors,
        array_key_1,
        array_key_2,
        affinity_array_key_1,
        affinity_array_key_2,
    ):
        self.array_key_1 = array_key_1
        self.array_key_2 = array_key_2
        self.affinity_array_key_1 = affinity_array_key_1
        self.affinity_array_key_2 = affinity_array_key_2
        self.affinity_vectors = affinity_vectors

    def setup(self):
        assert self.array_key_1 in self.spec, (
            "Upstream does not provide %s needed by \
        AddNonsymmetricAffinities"
            % self.array_key_1
        )
        assert self.array_key_2 in self.spec, (
            "Upstream does not provide %s needed by \
        AddNonsymmetricAffinities"
            % self.array_key_2
        )

        voxel_size = self.spec[self.array_key_1].voxel_size

        self.upstream_spec = self.get_upstream_provider().spec
        self.upstream_roi = self.upstream_spec.get_total_roi()

        # get maximum offset in each dimension
        self.padding = np.max(np.abs(self.affinity_vectors), axis=0)
        self.padding = tuple(round_to_voxel_size(self.padding, voxel_size))

        logger.debug("padding neg: %s" % np.asarray(self.padding))

        spec = self.spec[self.array_key_1].copy()
        # if spec.roi is not None:

        self.provides(self.affinity_array_key_1, spec)
        self.provides(self.affinity_array_key_2, spec)
        self.enable_autoskip()

    def prepare(self, request):
        array_1_roi = request[self.array_key_1].roi
        logger.debug("downstream %s request: " % self.array_key_1 + str(array_1_roi))

        array_2_roi = request[self.array_key_2].roi
        logger.debug("downstream %s request: " % self.array_key_2 + str(array_2_roi))

        # grow labels ROI to accomodate padding TODO: vol 2
        array_1_roi = array_1_roi.grow(self.padding, self.padding)
        array_2_roi = array_2_roi.grow(self.padding, self.padding)

        request[self.array_key_1].roi = array_1_roi
        request[self.array_key_2].roi = array_2_roi

        logger.debug("upstream %s request: " % self.array_key_1 + str(array_1_roi))
        logger.debug("upstream %s request: " % self.array_key_2 + str(array_2_roi))

        # pdb.set_trace()

    def process(self, batch, request):
        full_vol1 = batch.arrays[self.array_key_1]
        full_vol2 = batch.arrays[self.array_key_2]

        # Both full_vol1 should match
        assert (
            full_vol1.spec.dtype == full_vol2.spec.dtype
        ), "data type of array 1(%s) and array 2(%s) should match" % (
            full_vol1.spec.dtype,
            full_vol2.spec.dtype,
        )

        assert (
            full_vol1.spec.voxel_size == full_vol2.spec.voxel_size
        ), "data type of array 1(%s) and array 2(%s) should match" % (
            full_vol1.spec.voxel_size,
            full_vol2.spec.voxel_size,
        )

        logger.debug("computing ground-truth affinities from labels")

        # Calculate affinities 1: from vol2 onto vol1

        # Initialize affinity map
        request_vol = request[self.affinity_array_key_1]
        affinity_map = np.zeros(
            (len(self.affinity_vectors),)
            + tuple(request_vol.roi.shape / request_vol.voxel_size),
            dtype=full_vol1.spec.dtype,
        )

        # calculate affinities
        vol1 = full_vol1.crop(request_vol.roi)
        for i, vector in enumerate(self.affinity_vectors):
            vol2 = full_vol2.crop(request_vol.roi.shift(tuple(-vector)))
            affinity_map[i, :, :, :] = np.bitwise_and(vol1.data, vol2.data)

        batch.arrays[self.affinity_array_key_1] = Array(
            affinity_map, spec=request[self.affinity_array_key_1].copy()
        )

        batch.arrays[self.affinity_array_key_1].attrs[
            "affinity_vectors"
        ] = self.affinity_vectors

        # Calculate affinities 2: from vol1 onto vol2

        # Initialize affinity map
        request_vol = request[self.affinity_array_key_2]
        affinity_map = np.zeros(
            (len(self.affinity_vectors),)
            + tuple(request_vol.roi.shape / request_vol.voxel_size),
            dtype=full_vol1.spec.dtype,
        )

        # calculate affinities
        vol2 = full_vol2.crop(request_vol.roi)
        for i, vector in enumerate(self.affinity_vectors):
            vol1 = full_vol1.crop(request_vol.roi.shift(tuple(vector)))
            affinity_map[i, :, :, :] = np.bitwise_and(vol1.data, vol2.data)

        batch.arrays[self.affinity_array_key_2] = Array(
            affinity_map, spec=request[self.affinity_array_key_2].copy()
        )

        batch.arrays[self.affinity_array_key_2].attrs[
            "affinity_vectors"
        ] = self.affinity_vectors

        # Crop all other requests
        for array_key, array in request.array_specs.items():
            batch.arrays[array_key] = batch.arrays[array_key].crop(array.roi)

        for points_key, points in request.points_specs.items():
            recropped = batch.points[points_key].spec.roi = points.roi
            batch.points[points_key] = recropped


def round_to_voxel_size(shape, voxel_size):
    voxel_size = np.asarray(voxel_size, dtype=float)
    shape = np.ceil(shape / voxel_size) * voxel_size
    return np.array(shape, dtype="int32")



import copy

import numpy as np

from .freezable import Freezable


class GraphSpec(Freezable):
    """Contains meta-information about a graph. This is used by
    :class:`BatchProviders<BatchProvider>` to communicate the graphs they
    offer, as well as by :class:`Graph` to describe the data they contain.

    Attributes:

        roi (:class:`Roi`):

            The region of interested represented by this graph.

        directed (``bool``, optional):

            Whether the graph is directed or not.

        dtype (``dtype``, optional):

            The data type of the "location" attribute.
            Currently only supports np.float32.
    """

    def __init__(self, roi=None, directed=None, dtype=np.float32, placeholder=False):
        self.roi = roi
        self.directed = directed
        self.dtype = dtype
        self.placeholder = placeholder

        self.freeze()

    def update_with(self, spec):
        if self.roi is not None and spec.roi is not None:
            self.roi = self.roi.union(spec.roi)
        elif spec.roi is not None:
            self.roi = spec.roi

        if spec.directed is not None:
            self.directed = spec.directed

        if spec.dtype is not None:
            self.dtype = spec.dtype

        if spec.placeholder is not None:
            self.placeholder = spec.placeholder

    def copy(self):
        """Create a copy of this spec."""
        return copy.deepcopy(self)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __repr__(self):
        r = ""
        r += "ROI: " + str(self.roi) + ", "
        r += "dtype: " + str(self.dtype) + ", "
        r += "directed: " + str(self.directed) + ", "
        r += "placeholder: " + str(self.placeholder)
        return r
